from sklearn import linear_model
from sklearn.metrics import mean_squared_error as mse
from concurrent.futures import ThreadPoolExecutor
from pomegranate import DiscreteDistribution, ConditionalProbabilityTable
from bamt.utils.MathUtils import *
from gmr import GMM
# from sklearn.mixture import GaussianMixture
from pandas import DataFrame

from typing import Dict, List, Any, Union, TypedDict, Type, Optional, Sequence

import itertools
import random
import joblib
import pickle
import os

from bamt.log import logger_nodes
from bamt.config import config

STORAGE = config.get('NODES', 'models_storage', fallback='models_storage is not defined')


class DiscreteParams(TypedDict):
    cprob: Union[List[Union[list, Any]], Dict[str, list]]
    vals: List[str]


class BaseNode(object):
    """
    Base class for nodes.
    """

    def __init__(self, name: str):
        """
        :param name: name for node (taken from column name)
        type: node type
        disc_parents: list with discrete parents
        cont_parents: list with continuous parents
        children: node's children
        """
        self.name = name
        self.type = 'abstract'

        self.disc_parents = []
        self.cont_parents = []
        self.children = []

    def __repr__(self):
        return f"{self.name}"

    @staticmethod
    def choose_serialization(model) -> Union[str, Exception]:
        try:
            ex_b = pickle.dumps(model, protocol=4)
            model_ser = ex_b.decode('latin1').replace('\'', '\"')
            a = model_ser.replace('\"', '\'').encode('latin1')
            classifier_body = pickle.loads(a)
            return 'pickle'
        except Exception as ex:
            return ex


class DiscreteNode(BaseNode):
    """
    Main class of Discrete Node
    """

    def __init__(self, name):
        super(DiscreteNode, self).__init__(name)
        self.type = 'Discrete'

    def fit_parameters(self, data: DataFrame, num_workers: int = 1):
        """
        Train params for Discrete Node
        data: DataFrame to train on
        num_workers: number of Parallel Workers
        Method returns probas dict with following format {[<combinations>: value]}
        and vals, list of appeared values in combinations
        """

        def worker(node: Type[BaseNode]) -> DiscreteParams:
            parents = node.disc_parents + node.cont_parents
            if not parents:
                dist = DiscreteDistribution.from_samples(data[node.name].values)
                cprob = list(dict(sorted(dist.items())).values())
                vals = sorted([str(x) for x in list(dist.parameters[0].keys())])
            else:
                dist = DiscreteDistribution.from_samples(data[node.name].values)
                vals = sorted([str(x) for x in list(dist.parameters[0].keys())])
                dist = ConditionalProbabilityTable.from_samples(data[parents + [node.name]].values)
                params = dist.parameters[0]
                cprob = dict()
                for i in range(0, len(params), len(vals)):
                    probs = []
                    for j in range(i, (i + len(vals))):
                        probs.append(params[j][-1])
                    combination = [str(x) for x in params[i][0:len(parents)]]
                    cprob[str(combination)] = probs
            return {"cprob": cprob, 'vals': vals}

        pool = ThreadPoolExecutor(num_workers)
        future = pool.submit(worker, self)
        return future.result()

    @staticmethod
    def choose(node_info: Dict[str, Union[float, str]], pvals: List[str]) -> str:
        """
        Return value from discrete node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """
        rindex = 0
        random.seed()
        vals = node_info['vals']
        if not pvals:
            dist = node_info['cprob']
        else:
            # noinspection PyTypeChecker
            dist = node_info['cprob'][str(pvals)]
        lbound = 0
        ubound = 0
        rand = random.random()
        for interval in range(len(dist)):
            ubound += dist[interval]
            if lbound <= rand < ubound:
                rindex = interval
                break
            else:
                lbound = ubound

        return vals[rindex]

    @staticmethod
    def predict(node_info: Dict[str, Union[float, str]], pvals: List[str]) -> str:
        """function for prediction based on evidence values in discrete node

        Args:
            node_info (Dict[str, Union[float, str]]): parameters of node
            pvals (List[str]): values in parents nodes

        Returns:
            str: prediction
        """        
        # vals = node_info['vals']
        # if not pvals:
        #     dist = node_info['cprob']
        # else:
        #     # noinspection PyTypeChecker
        #     dist = node_info['cprob'][str(pvals)]
        # index_max = 0    
        # g = itertools.groupby(dist)
        # if next(g, True) and not next(g, False):
        #     index_max = random.randint(0, len(dist))
        # else:
        #     index_max = np.argmax(dist)
        # return vals[index_max]

        vals = node_info['vals']
        disct = []
        if not pvals:
            dist = node_info['cprob']
        else:
            # noinspection PyTypeChecker
            dist = node_info['cprob'][str(pvals)]
        max_value = max(dist)
        indices = [index for index, value in enumerate(dist) if value == max_value]
        max_ind = 0
        if len(indices) == 1:
            max_ind = indices[0]
        else:
            max_ind = random.choice(indices)
        return vals[max_ind]


class GaussianParams(TypedDict):
    mean: np.ndarray
    coef: Sequence[Any]
    variance: Union[np.ndarray, float]


class GaussianNode(BaseNode):
    """
    Main class for Gaussian Node
    """

    def __init__(self, name: str, model=linear_model.LinearRegression()):
        super(GaussianNode, self).__init__(name)
        self.model = model
        self.type = 'Gaussian'

    def fit_parameters(self, data: DataFrame) -> GaussianParams:
        """
        Function for training parameters for gaussian node
        """
        parents = self.disc_parents + self.cont_parents
        if parents:
            # model = self.model
            predict = []
            if len(parents) == 1:
                self.model.fit(np.transpose([data[parents[0]].values]), data[self.name].values)
                predict = self.model.predict(np.transpose([data[parents[0]].values]))
            else:
                self.model.fit(data[parents].values, data[self.name].values)
                predict = self.model.predict(data[parents].values)
            variance = mse(data[self.name].values, predict)
            return {"mean": self.model.intercept_,
                    "coef": list(self.model.coef_),
                    "variance": variance}
        else:
            mean_base = np.mean(data[self.name].values)
            self.model.intercept_ = mean_base
            self.model.coef_ = np.array([])
            variance = np.var(data[self.name].values)
            return {"mean": mean_base,
                    "coef": [],
                    "variance": variance}

    @staticmethod
    def choose(node_info: Dict[str, Union[float, List[float]]],
               pvals: List[float]) -> float:
        """
        Return value from Gaussian node
        params:
        node_info: information about node
        pvals: parent values
        """
        mean = node_info["mean"]
        if pvals:
            for i, m in enumerate(pvals):
                mean += m * node_info['coef'][i]
        variance = node_info['variance']
        # distribution = [mean, variance]
        return random.gauss(mean, math.sqrt(variance))

    @staticmethod
    def predict(node_info: Dict[str, Union[float, List[float]]],
                pvals: List[float]) -> float:
        """function for prediction in gaussian node

        Args:
            node_info (Dict[str, Union[float, List[float]]]): node parameters
            pvals (List[float]): parent values

        Returns:
            float: prediction
        """
        mean = node_info["mean"]
        if pvals:
            for i, m in enumerate(pvals):
                mean += m * node_info['coef'][i]
        return mean


class CondGaussParams(TypedDict):
    variance: Optional[float]
    mean: Optional[List[float]]
    coef: List[float]


class ConditionalGaussianNode(BaseNode):
    """
    Main class for Conditional Gaussian Node
    """

    def __init__(self, name):
        super(ConditionalGaussianNode, self).__init__(name)
        self.type = 'ConditionalGaussian'

    def fit_parameters(self, data: DataFrame) -> Dict[str, Dict[str, CondGaussParams]]:
        """
        Train params for Conditional Gaussian Node.
        Return:
        {"hybcprob": {<combination of outputs from discrete parents> : CondGaussParams}}
        """
        hycprob = dict()
        values = []
        combinations = []
        for d_p in self.disc_parents:
            values.append(np.unique(data[d_p].values))
        for xs in itertools.product(*values):
            combinations.append(list(xs))
        for comb in combinations:
            mask = np.full(len(data), True)
            for col, val in zip(self.disc_parents, comb):
                mask = (mask) & (data[col] == val)
            new_data = data[mask]
            mean_base = np.nan
            variance = np.nan
            key_comb = [str(x) for x in comb]
            if new_data.shape[0] != 0:
                if self.cont_parents:
                    model = linear_model.LinearRegression()
                    if len(self.cont_parents) == 1:
                        model.fit(np.transpose([new_data[self.cont_parents[0]].values]), new_data[self.name].values)
                        predict = model.predict(np.transpose([new_data[self.cont_parents[0]].values]))
                    else:
                        model.fit(new_data[self.cont_parents].values, new_data[self.name].values)
                        predict = model.predict(new_data[self.cont_parents].values)
                    mean_base = model.intercept_
                    variance = mse(new_data[self.name].values, predict)
                    hycprob[str(key_comb)] = {'variance': variance, 'mean': mean_base,
                                              'coef': list(model.coef_)}
                else:
                    mean_base = np.mean(new_data[self.name].values)
                    variance = np.var(new_data[self.name].values)
                    hycprob[str(key_comb)] = {'variance': variance, 'mean': mean_base, 'coef': []}
            else:
                if self.cont_parents:
                    scal = list(np.full(len(self.cont_parents), np.nan))
                    hycprob[str(key_comb)] = {'variance': variance, 'mean': mean_base, 'coef': scal}
                else:
                    # mean_base = np.nan
                    # variance = np.nan
                    hycprob[str(key_comb)] = {'variance': variance, 'mean': mean_base, 'coef': []}
        return {"hybcprob": hycprob}

    @staticmethod
    def choose(node_info: Dict[str, Dict[str, CondGaussParams]], pvals: List[Union[str, float]]) -> float:
        """
        Return value from ConditionalGaussian node
        node_info: nodes info from distributions
        pvals: parent values
        """
        dispvals = []
        lgpvals = []
        for pval in pvals:
            if (isinstance(pval, str)) | (isinstance(pval, int)):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)
        lgdistribution = node_info["hybcprob"][str(dispvals)]
        mean = lgdistribution["mean"]
        if lgpvals:
            for x in range(len(lgpvals)):
                mean += lgpvals[x] * lgdistribution["coef"][x]
        variance = lgdistribution["variance"]
        return random.gauss(mean, math.sqrt(variance))

    @staticmethod
    def predict(node_info: Dict[str, Dict[str, CondGaussParams]], pvals: List[Union[str, float]]) -> float:
        """function for prediction in conditional gaussian node

        Args:
            node_info (Dict[str, Union[float, List[float]]]): node parameters
            pvals (List[float]): parent values

        Returns:
            float: prediction
        """
        dispvals = []
        lgpvals = []
        for pval in pvals:
            if (isinstance(pval, str)) | (isinstance(pval, int)):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)
        lgdistribution = node_info["hybcprob"][str(dispvals)]
        mean = lgdistribution["mean"]
        if lgpvals:
            for x in range(len(lgpvals)):
                mean += lgpvals[x] * lgdistribution["coef"][x]
        return mean


class MixtureGaussianParams(TypedDict):
    mean: List[float]
    coef: List[float]
    covars: List[float]


class MixtureGaussianNode(BaseNode):
    """
    Main class for Mixture Gaussian Node
    """

    def __init__(self, name):
        super(MixtureGaussianNode, self).__init__(name)
        self.type = 'MixtureGaussian'

    def fit_parameters(self, data: DataFrame) -> MixtureGaussianParams:
        """
        Train params for Mixture Gaussian Node
        """
        parents = self.disc_parents + self.cont_parents
        if not parents:
            n_comp = int((component(data, [self.name], 'aic') + component(data, [self.name],
                                                                          'bic')) / 2)  # component(data, [node], 'LRTS')#
            # n_comp = 3
            gmm = GMM(n_components=n_comp).from_samples(np.transpose([data[self.name].values]), n_iter=500,
                                                        init_params='kmeans++')
            means = gmm.means.tolist()
            cov = gmm.covariances.tolist()
            # weigts = np.transpose(gmm.to_responsibilities(np.transpose([data[node].values])))
            w = gmm.priors.tolist()  # []
            # for row in weigts:
            #     w.append(np.mean(row))
            return {"mean": means, "coef": w, "covars": cov}
        if parents:
            if not self.disc_parents and self.cont_parents:
                nodes = [self.name] + self.cont_parents
                new_data = data[nodes]
                new_data.reset_index(inplace=True, drop=True)
                n_comp = int((component(new_data, nodes, 'aic') + component(new_data, nodes,
                                                                            'bic')) / 2)  # component(new_data, nodes, 'LRTS')#
                # n_comp = 3
                gmm = GMM(n_components=n_comp).from_samples(new_data[nodes].values, n_iter=500, init_params='kmeans++')
                means = gmm.means.tolist()
                cov = gmm.covariances.tolist()
                # weigts = np.transpose(gmm.to_responsibilities(new_data[nodes].values))
                w = gmm.priors.tolist()  # []
                # for row in weigts:
                #     w.append(np.mean(row))
                return {"mean": means,
                        "coef": w,
                        "covars": cov}

    @staticmethod
    def choose(node_info: MixtureGaussianParams, pvals: List[Union[str, float]]) -> Optional[float]:
        """
        Func to get value from current node
        node_info: nodes info from distributions
        pvals: parent values
        Return value from MixtureGaussian node
        """
        mean = node_info["mean"]
        covariance = node_info["covars"]
        w = node_info["coef"]
        n_comp = len(node_info['coef'])
        if n_comp != 0:
            if pvals:
                indexes = [i for i in range(1, len(pvals) + 1)]
                if not np.isnan(np.array(pvals)).all():
                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                    cond_gmm = gmm.condition(indexes, [pvals])
                    sample = cond_gmm.sample(1)[0][0]
                else:
                    sample = np.nan
            else:
                gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                sample = gmm.sample(1)[0][0]
        else:
            sample = np.nan
        return sample

    @staticmethod
    def predict(node_info: MixtureGaussianParams, pvals: List[Union[str, float]]) -> Optional[float]:
        """
        Func to get prediction from current node
        node_info: nodes info from distributions
        pvals: parent values
        Return value from MixtureGaussian node
        """
        mean = node_info["mean"]
        covariance = node_info["covars"]
        w = node_info["coef"]
        n_comp = len(node_info['coef'])
        if n_comp != 0:
            if pvals:
                indexes = [i for i in range(1, len(pvals) + 1)]
                if not np.isnan(np.array(pvals)).all():
                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                    sample = gmm.predict(indexes, [pvals])[0][0]
                else:
                    sample = np.nan
            else:
                # gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                sample = 0
                for ind, wi in enumerate(w):
                    sample += wi * mean[ind][0]
        else:
            sample = np.nan
        return sample


class CondMixtureGaussParams(TypedDict):
    mean: Optional[List[float]]
    coef: List[float]
    covars: Optional[List[float]]


class ConditionalMixtureGaussianNode(BaseNode):
    """
    Main class for Conditional Mixture Gaussian Node
    """

    def __init__(self, name):
        super(ConditionalMixtureGaussianNode, self).__init__(name)
        self.type = 'ConditionalMixtureGaussian'

    def fit_parameters(self, data: DataFrame) -> Dict[str, Dict[str, CondMixtureGaussParams]]:
        """
        Train params for Conditional Mixture Gaussian Node.
        Return:
        {"hybcprob": {<combination of outputs from discrete parents> : CondMixtureGaussParams}}
        """
        hycprob = dict()
        values = []
        combinations = []
        for d_p in self.disc_parents:
            values.append(np.unique(data[d_p].values))
        for xs in itertools.product(*values):
            combinations.append(list(xs))
        for comb in combinations:
            mask = np.full(len(data), True)
            for col, val in zip(self.disc_parents, comb):
                mask = (mask) & (data[col] == val)
            new_data = data[mask]
            new_data.reset_index(inplace=True, drop=True)
            key_comb = [str(x) for x in comb]
            nodes = [self.name] + self.cont_parents
            if new_data.shape[0] > 5:
                if self.cont_parents:
                    n_comp = int((component(new_data, nodes, 'aic') + component(new_data, nodes,
                                                                                'bic')) / 2)  # component(new_data, nodes, 'LRTS')#int((component(new_data, nodes, 'aic') + component(new_data, nodes, 'bic')) / 2)
                    # n_comp = 3
                    gmm = GMM(n_components=n_comp).from_samples(new_data[nodes].values, n_iter=500,
                                                                init_params='kmeans++')
                else:
                    n_comp = int((component(new_data, [self.name], 'aic') + component(new_data, [self.name],
                                                                                      'bic')) / 2)  # component(new_data, [node], 'LRTS')#int((component(new_data, [node], 'aic') + component(new_data, [node], 'bic')) / 2)
                    # n_comp = 3
                    gmm = GMM(n_components=n_comp).from_samples(np.transpose([new_data[self.name].values]), n_iter=500,
                                                                init_params='kmeans++')
                means = gmm.means.tolist()
                cov = gmm.covariances.tolist()
                # weigts = np.transpose(gmm.to_responsibilities(np.transpose([new_data[node].values])))
                w = gmm.priors.tolist()  # []
                # for row in weigts:
                #     w.append(np.mean(row))
                hycprob[str(key_comb)] = {'covars': cov, 'mean': means, 'coef': w}
            elif new_data.shape[0] != 0:
                n_comp = 1
                gmm = GMM(n_components=n_comp)
                if self.cont_parents:
                    gmm.from_samples(new_data[nodes].values)
                else:
                    gmm.from_samples(np.transpose([new_data[self.name].values]))
                means = gmm.means.tolist()
                cov = gmm.covariances.tolist()
                # weigts = np.transpose(gmm.to_responsibilities(np.transpose([new_data[node].values])))
                w = gmm.priors.tolist()  # []
                # for row in weigts:
                #     w.append(np.mean(row))
                hycprob[str(key_comb)] = {'covars': cov, 'mean': means, 'coef': w}
            else:
                if self.cont_parents:
                    hycprob[str(key_comb)] = {'covars': np.nan, 'mean': np.nan, 'coef': []}
                else:
                    hycprob[str(key_comb)] = {'covars': np.nan, 'mean': np.nan, 'coef': []}
        return {"hybcprob": hycprob}

    @staticmethod
    def choose(node_info: Dict[str, Dict[str, CondMixtureGaussParams]],
               pvals: List[Union[str, float]]) -> Optional[float]:
        """
        Function to get value from ConditionalMixtureGaussian node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """
        dispvals = []
        lgpvals = []
        for pval in pvals:
            if ((isinstance(pval, str)) | ((isinstance(pval, int)))):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)
        lgdistribution = node_info["hybcprob"][str(dispvals)]
        mean = lgdistribution["mean"]
        covariance = lgdistribution["covars"]
        w = lgdistribution["coef"]
        if len(w) != 0:
            if len(lgpvals) != 0:
                indexes = [i for i in range(1, (len(lgpvals) + 1), 1)]
                if not np.isnan(np.array(lgpvals)).all():
                    n_comp = len(w)
                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                    cond_gmm = gmm.condition(indexes, [lgpvals])
                    sample = cond_gmm.sample(1)[0][0]
                else:
                    sample = np.nan
            else:
                n_comp = len(w)
                gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                sample = gmm.sample(1)[0][0]
        else:
            sample = np.nan
        return sample

    @staticmethod
    def predict(node_info: Dict[str, Dict[str, CondMixtureGaussParams]],
                pvals: List[Union[str, float]]) -> Optional[float]:
        """
        Function to get prediction from ConditionalMixtureGaussian node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        dispvals = []
        lgpvals = []
        for pval in pvals:
            if ((isinstance(pval, str)) | ((isinstance(pval, int)))):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)
        lgdistribution = node_info["hybcprob"][str(dispvals)]
        mean = lgdistribution["mean"]
        covariance = lgdistribution["covars"]
        w = lgdistribution["coef"]
        if len(w) != 0:
            if len(lgpvals) != 0:
                indexes = [i for i in range(1, (len(lgpvals) + 1), 1)]
                if not np.isnan(np.array(lgpvals)).all():
                    n_comp = len(w)
                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                    sample = gmm.predict(indexes, [lgpvals])[0][0]
                else:
                    sample = np.nan
            else:
                # n_comp = len(w)
                # gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=covariance)
                sample = 0
                for ind, wi in enumerate(w):
                    sample += wi * mean[ind][0]
        else:
            sample = np.nan
        return sample


class LogitParams(TypedDict):
    classes: List[int]
    classifier: str
    classifier_obj: Optional[Union[str, bool, bytes]]
    serialization: str


class LogitNode(BaseNode):
    """
    Main class for logit node
    """

    def __init__(self, name, classifier: Optional[object] = None):
        super(LogitNode, self).__init__(name)
        if classifier is None:
            classifier = linear_model.LogisticRegression(multi_class='multinomial', solver='newton-cg', max_iter=100)
        self.classifier = classifier
        self.type = 'Logit' + f" ({type(self.classifier).__name__})"

    def fit_parameters(self, data: DataFrame) -> LogitParams:
        parents = self.disc_parents + self.cont_parents
        self.classifier.fit(data[parents].values, data[self.name].values)
        serialization = self.choose_serialization(self.classifier)

        if serialization == 'pickle':
            ex_b = pickle.dumps(self.classifier, protocol=4)
            # model_ser = ex_b.decode('latin1').replace('\'', '\"')
            model_ser = ex_b.decode('latin1')
            return {'classes': list(self.classifier.classes_),
                    'classifier_obj': model_ser,
                    'classifier': type(self.classifier).__name__,
                    'serialization': 'pickle'}
        else:
            logger_nodes.warning(f"{self.name}::Pickle failed. BAMT will use Joblib. | " + str(serialization.args[0]))
            index = str(int(os.listdir(STORAGE)[-1]))
            if not os.path.isdir(os.path.join(STORAGE, index, f"{self.name.replace(' ', '_')}")):
                os.makedirs(os.path.join(STORAGE, index, f"{self.name.replace(' ', '_')}"))
            path = os.path.abspath(os.path.join(STORAGE,
                                                index,
                                                f"{self.name.replace(' ', '_')}",
                                                f"{self.name.replace(' ', '_')}.joblib.compressed"))

            joblib.dump(self.classifier, path, compress=True, protocol=4)
            return {'classes': list(self.classifier.classes_),
                    'classifier_obj': path,
                    'classifier': type(self.classifier).__name__,
                    'serialization': 'joblib'}

    def choose(self, node_info: LogitParams, pvals: List[Union[str, float]]) -> str:
        """
        Return value from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """
        pvals = [str(p) for p in pvals]

        rindex = 0

        if len(node_info["classes"]) > 1:
            if node_info["serialization"] == 'joblib':
                model = joblib.load(node_info["classifier_obj"])
            else:
                # str_model = node_info["classifier_obj"].decode('latin1').replace('\'', '\"')
                a = node_info["classifier_obj"].encode('latin1')
                model = pickle.loads(a)

            distribution = model.predict_proba(np.array(pvals).reshape(1, -1))[0]

            # choose
            rand = random.random()
            lbound = 0
            ubound = 0
            for interval in range(len(node_info["classes"])):
                ubound += distribution[interval]
                if (lbound <= rand and rand < ubound):
                    rindex = interval
                    break
                else:
                    lbound = ubound

            return str(node_info["classes"][rindex])

        else:
            return str(node_info["classes"][0])

    def predict(self, node_info: LogitParams, pvals: List[Union[str, float]]) -> str:
        """
        Return prediction from Logit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """
        pvals = [str(p) for p in pvals]

        if len(node_info["classes"]) > 1:
            if node_info["serialization"] == 'joblib':
                model = joblib.load(node_info["classifier_obj"])
            else:
                # str_model = node_info["classifier_obj"].decode('latin1').replace('\'', '\"')
                a = node_info["classifier_obj"].encode('latin1')
                model = pickle.loads(a)

            pred = model.predict(np.array(pvals).reshape(1, -1))[0]

            return str(pred)

        else:
            return str(node_info["classes"][0])


class ConditionalLogitNode(BaseNode):
    """
    Main class for Conditional Logit Node
    """

    def __init__(self, name: str, classifier: Optional[object] = None):
        super(ConditionalLogitNode, self).__init__(name)
        if classifier is None:
            classifier = linear_model.LogisticRegression(multi_class='multinomial', solver='newton-cg', max_iter=100)
        self.classifier = classifier
        self.type = 'ConditionalLogit' + f" ({type(self.classifier).__name__})"

    def fit_parameters(self, data: DataFrame) -> Dict[str, Dict[str, LogitParams]]:
        """
        Train params on data
        Return:
        {"hybcprob": {<combination of outputs from discrete parents> : LogitParams}}
        """
        hycprob = dict()
        values = []
        combinations = []
        for d_p in self.disc_parents:
            values.append(np.unique(data[d_p].values))
        for xs in itertools.product(*values):
            combinations.append(list(xs))
        for comb in combinations:
            mask = np.full(len(data), True)
            for col, val in zip(self.disc_parents, comb):
                mask = (mask) & (data[col] == val)
            new_data = data[mask]
            # mean_base = [np.nan]
            classes = []
            key_comb = [str(x) for x in comb]
            if new_data.shape[0] != 0:
                model = self.classifier
                values = set(new_data[self.name])
                if len(values) > 1:
                    model.fit(new_data[self.cont_parents].values, new_data[self.name].values)
                    classes = list(model.classes_)
                    serialization = self.choose_serialization(model)

                    if serialization == 'pickle':
                        ex_b = pickle.dumps(self.classifier, protocol=4)
                        model_ser = ex_b.decode('latin1')

                        # model_ser = pickle.dumps(self.classifier, protocol=4)
                        hycprob[str(key_comb)] = {'classes': classes,
                                                  'classifier_obj': model_ser,
                                                  'classifier': type(self.classifier).__name__,
                                                  'serialization': 'pickle'}
                    else:
                        logger_nodes.warning(
                            f"{self.name} {comb}::Pickle failed. BAMT will use Joblib. | " + str(serialization.args[0]))
                        index = str(int(os.listdir(STORAGE)[-1]))
                        if not os.path.isdir(os.path.join(STORAGE, index, f"{self.name.replace(' ', '_')}")):
                            os.makedirs(os.path.join(STORAGE, index, f"{self.name.replace(' ', '_')}"))
                        path = os.path.abspath(os.path.join(STORAGE,
                                                            index,
                                                            f"{self.name.replace(' ', '_')}",
                                                            f"{comb}.joblib.compressed"))

                        joblib.dump(model, path, compress=True, protocol=4)
                        hycprob[str(key_comb)] = {'classes': classes,
                                                  'classifier_obj': path,
                                                  'classifier': type(self.classifier).__name__,
                                                  'serialization': 'joblib'}
                else:
                    classes = list(values)
                    hycprob[str(key_comb)] = {'classes': classes, 'classifier': type(self.classifier).__name__,
                                              'classifier_obj': None, 'serialization': None}

            else:
                hycprob[str(key_comb)] = {'classes': list(classes), 'classifier': type(self.classifier).__name__,
                                          'classifier_obj': None, 'serialization': None}
        return {"hybcprob": hycprob}

    def choose(self, node_info: Dict[str, Dict[str, LogitParams]], pvals: List[Union[str, float]]) -> str:
        """
        Return value from ConditionalLogit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        dispvals = []
        lgpvals = []
        for pval in pvals:
            if (isinstance(pval, str)):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)

        lgdistribution = node_info["hybcprob"][str(dispvals)]

        # JOBLIB
        if len(lgdistribution["classes"]) > 1:
            if lgdistribution["serialization"] == 'joblib':
                model = joblib.load(lgdistribution["classifier_obj"])
            else:
                # str_model = lgdistribution["classifier_obj"].decode('latin1').replace('\'', '\"')
                bytes_model = lgdistribution["classifier_obj"].encode('latin1')
                model = pickle.loads(bytes_model)

            distribution = model.predict_proba(np.array(lgpvals).reshape(1, -1))[0]

            rand = random.random()
            rindex = 0
            lbound = 0
            ubound = 0
            for interval in range(len(lgdistribution["classes"])):
                ubound += distribution[interval]
                if (lbound <= rand and rand < ubound):
                    rindex = interval
                    break
                else:
                    lbound = ubound
            return str(lgdistribution["classes"][rindex])

        else:
            return str(lgdistribution["classes"][0])

    def predict(self, node_info: Dict[str, Dict[str, LogitParams]], pvals: List[Union[str, float]]) -> str:
        """
        Return value from ConditionalLogit node
        params:
        node_info: nodes info from distributions
        pvals: parent values
        """

        dispvals = []
        lgpvals = []
        for pval in pvals:
            if (isinstance(pval, str)):
                dispvals.append(pval)
            else:
                lgpvals.append(pval)

        lgdistribution = node_info["hybcprob"][str(dispvals)]

        # JOBLIB
        if len(lgdistribution["classes"]) > 1:
            if lgdistribution["serialization"] == 'joblib':
                model = joblib.load(lgdistribution["classifier_obj"])
            else:
                # str_model = lgdistribution["classifier_obj"].decode('latin1').replace('\'', '\"')
                bytes_model = lgdistribution["classifier_obj"].encode('latin1')
                model = pickle.loads(bytes_model)

            pred = model.predict(np.array(lgpvals).reshape(1, -1))[0]

            return str(pred)

        else:
            return str(lgdistribution["classes"][0])



"""
************
Factor Class
************

This class holds a Conditional Probability Table structure
-- i.e. a factor. The benefit of this class structure is that
all factor manipulation happens in a centralized location,
thereby making it easier to write fast and readable code.

The Joint Probability Distribution of a Bayesian Network is
simply a product of its factors. Much of the functionality
is derived from algorithms presented in [1].

The tests for this class are found in "test_factor.py".

For accessing the flattened array based on RV values/indices
and respective strides, use this formula:
sum( value_index[i]*stride[i] for i = all variables in the scope )


References
----------
[1] Koller, Friedman (2009). "Probabilistic Graphical Models."

"""
from __future__ import division

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np

class Factor(object):
    """
    A Factor uses a flattened numpy array for the cpt.
    By storing the cpt in this manner and taking advantage 
    of efficient algorithms, significant speedups occur.

    Attributes
    ----------

    *self.var* : a string
        The random variable to which this Factor belongs
    
    *self.scope* : a list
        The RV, and its parents (the RVs involved in the
        conditional probability table)
    
    *self.stride* : a dictionary, where
        key = an RV in self.scope, and
        val = integer stride (i.e. how many rows in the 
            CPT until the NEXT value of RV is reached)
    
    *self.cpt* : a 1D numpy array
        The probability values for self.var conditioned
        on its parents
    

    Methods
    -------
    *multiply_factor*
        Multiply two factors together. The factor
        multiplication algorithm used here is adapted
        from Koller and Friedman (PGMs) textbook.

    *sumover_var* :
        Sum over one *rv* by keeping it constant. Thus, you 
        end up with a 1-D factor whose scope is ONLY *rv*
        and whose length = cardinality of rv. 

    *sumout_var_list* :
        Remove a collection of rv's from the factor
        by summing out (i.e. calling sumout_var) over
        each rv.

    *sumout_var* :
        Remove passed-in *rv* from the factor by summing
        over everything else.

    *maxout_var* :
        Remove *rv* from the factor by taking the maximum value 
        of all rv instantiations over everyting else.

    *reduce_factor_by_list* :
        Reduce the factor by numerous sets of
        [rv,val]

    *reduce_factor* :
        Condition the factor by eliminating any sets of
        values that don't align with a given [rv, val]

    *to_log* :
        Convert probabilities to log space from
        normal space.

    *from_log* :
        Convert probabilities from log space to
        normal space.

    *normalize* :
        Make relevant collections of probabilities sum to one.


    Notes
    -----
    """            


    def __init__(self, bn, var):
        """
        Initialize a Factor from a BayesNet object
        for a given random variable.

        Note, it's assumed that the FIRST variable
        of *scope* is the main variable (i.e. NOT a
        parent).

        Arguments
        ---------

        *var* : a string
            The RV for which the Factor will be extracted.

        Effects
        -------
        - sets *self.var*
        - sets *self.cpt*
        - sets *self.card*
        - sets *self.scope*
        - sets *self.stride*

        Notes
        -----
        - self.card is no longer an attribute, but is now a function
        - self.bn is no longer an attribute

        """
        self.bn = bn
        self.var = var
        self.cpt = np.array(bn.cpt(var))
        self.scope = bn.scope(var)
        self.card = dict([(rv, bn.card(rv)) for rv in self.scope])

        self.stride = {self.var:1}
        s=self.card[self.var]
        for v in bn.parents(var):
            self.stride[v]=s
            s*=self.card[v]


    def __repr__(self):
        """
        Internal representation of the factor,
        to be used when the object is called
        in the console without print.
        """
        s = self.var + ' | '
        s += ', '.join(self.parents())
        return s

    def __str__(self):
        """
        String representation of the factor,
        to be used when print is called.
        """
        s = self.var + ' | '
        s += ', '.join(self.parents())
        return s

    def __mul__(self, other_factor):
        """
        Overloads multiplication operator to
        be used as multiplying two factors together.
        """
        self.multiply_factor(other_factor)
        return self

    def __sub__(self, rv_val):
        """
        Overloads subtraction operator to
        be used as reducing a factor by evidence.
        """
        self.reduce_factor(rv_val[0],rv_val[1])
        return self

    def __div__(self, rv):
        """
        Overloads division operator to
        be used as summing out a variable.
        """
        self.sumout_var(rv)
        return self

    def __floordiv__(self, rv):
        """
        Overloads floor division operator to
        be used as maxing out a variable
        """
        self.maxout_var(rv)
        return self


    def parents(self):
        """
        Return parents of self.var ...
        Should make this an iterator
        """
        for rv in self.scope:
            if rv != self.var:
                yield rv
                
    def values(self, rv):
        return self.bn.values(rv)

    def value_indices(self, val_dict):
        """
        Return the indices in the cpt
        where RV=Value in val_dict
        For accessing the flattened array based 
        on RV values/indices
        and respective strides, use this formula:
        sum( value_index[i]*stride[i] for i = all 
            variables in the scope )
        """
        idx = sum([self.bn.value_idx(rv,val)*self.stride[rv] \
            for rv,val in val_dict.items()])
        return idx

    def sepset(self, other_factor):
        """
        The sepset of two cliques is the set of
        variables in the intersection of the two
        cliques' scopes.

        Arguments
        ---------
        *other_clique* : a Clique object
        """
        return set(self.scope).intersection(set(other_factor.scope))



    ##### FACTOR OPERATIONS #####

    def multiply_factor(self, other_factor):
        """
        Multiply two factors together. The factor
        multiplication algorithm used here is adapted
        from Koller and Friedman (PGMs) textbook.

        In essence, the scope of the merged factor is the
        union of the two scopes.

        Arguments
        ---------
        *other_factor* : a different Factor object

        Returns
        -------
        None

        Effects
        -------
        - alters self.cpt
        - alters self.stride
        - alters self.card
        - alters self.scope

        Notes
        -----
        - What is done about normalization here? I guess
        assume it's already normalized

        """
        if len(self.scope)>=len(other_factor.scope):
            phi1=self
            phi2=other_factor
        else:
            phi1=other_factor
            phi2=self
        # go in order of strides to keep them in order after the fact
        rv_order = sorted(phi1.stride, key=phi1.stride.__getitem__)        
        scope_set=set(phi1.scope).union(set(phi2.scope))

        ########### HAD TO ADD THESE FOR V/E TO WORK... #################
        ## I think this is necessary when phi2 has vars not in phi1 ##
        rv_order.extend(list(set(phi2.scope).difference(set(phi1.scope))))
        phi1.card.update(phi2.card)
        #################################################################

        j,k=0,0
        assignment = dict([(rv, 0) for rv in scope_set])
        psi = np.zeros(np.product(phi1.card.values()))

        for i in range(len(psi)):
            psi[i] = phi1.cpt[j]*phi2.cpt[k]
            for rv in rv_order:
                assignment[rv] += 1
                if assignment[rv] == phi1.card[rv]:
                    assignment[rv] = 0
                    if rv in phi1.scope:
                        j=j-(phi1.card[rv]-1)*phi1.stride[rv]
                    if rv in phi2.scope:
                        k=k-(phi2.card[rv]-1)*phi2.stride[rv]
                else:
                    if rv in phi1.scope:
                        j=j+phi1.stride[rv]
                    if rv in phi2.scope:
                        k=k+phi2.stride[rv]
                    break
        self.cpt = psi
        self.scope = list(scope_set)
        self.card=phi1.card
        self.stride={}
        s=1
        for v in rv_order:
            self.stride[v]=s
            s*=phi1.card[v]

        #self.normalize()


    def sumover_var(self, rv):
        """
        Sum over one *rv* by keeping it constant. Thus, you 
        end up with a factor whose scope is ONLY *rv*
        and whose length = cardinality of rv. 

        This is equivalent to calling self.sumout_var() over
        EVERY other variable in the scope and is thus faster
        when you want to do just that.

        Arguments
        ---------
        *rv* : a string
            The random variable to sum over.

        Returns
        -------
        None

        Effects
        -------
        - alters self.cpt
        - alters self.stride
        - alters self.card
        - alters self.scope

        Notes
        -----

        """
        exp_len = self.card[rv]
        new_cpt= np.zeros(exp_len)

        rv_card = self.card[rv]
        rv_stride = self.stride[rv]

        for i in range(exp_len):
            idx=i*rv_stride
            while idx < len(self.cpt):
                new_cpt[i] += np.sum(self.cpt[idx:(idx+rv_stride)])
                idx+=rv_card*rv_stride

        self.cpt=new_cpt
        self.card = {rv:rv_card}
        self.stride = {rv:1}
        self.scope = [rv]
        self.var = rv

        #self.normalize()

    def sumout_var_list(self, var_list):
        """
        Remove a collection of rv's from the factor
        by summing out (i.e. calling sumout_var) over
        each rv.

        Arguments
        ---------
        *var_list* : a list
            The list of rv's to sum out.

        Returns
        -------
        None

        Effects
        -------
        - see "self.sumout_var"

        Notes
        -----

        """
        for var in var_list:
            self.sumout_var(var)

    def sumout_var(self, rv):
        """
        Remove passed-in *rv* from the factor by summing
        over everything else.

        Arguments
        ---------
        *rv* : a string
            The random variable to sum out

        Returns
        -------
        None

        Effects
        -------
        - alters self.cpt
        - alters self.stride
        - alters self.card
        - alters self.scope

        Notes
        -----     
        
        """
        exp_len = int(len(self.cpt)/self.card[rv])
        new_cpt = np.zeros(exp_len)
        
        rv_card = self.card[rv]
        rv_stride = self.stride[rv]

        k=0
        p = np.prod([self.card[r] for r in self.scope if self.stride[r] < self.stride[rv]])
        for i in range(exp_len):
            for c in range(rv_card):
                new_cpt[i] += self.cpt[k + (rv_stride*c)]
            k+=1
            if (k % p== 0):
                k += (p * (rv_card - 1))
        self.cpt=new_cpt

        del self.card[rv]
        self.stride.update((k,v/rv_card) for k,v in self.stride.items() if v > rv_stride)
        del self.stride[rv]
        self.scope.remove(rv)

        if rv == self.var:
            l = [k for k,v in self.stride.items() if v==1]
            if len(l)>0:
                self.var = l[0]

        #if len([k for k,v in self.stride.items() if v==1]) > 0:
        #self.normalize()

    def maxout_var(self, rv):
        """
        Remove *rv* from the factor by taking the maximum value 
        of all instantiations of the passed-in rv

        Used in MAP inference (i.e. Algorithm 13.1 in Koller p.557)

        Arguments
        ---------
        *rv* : a string
            The random variable

        Returns
        -------
        None

        Effects
        -------
        - alters self.cpt
        - alters self.stride
        - alters self.card
        - alters self.scope

        Notes
        -----        
        
        """
        #self.cpt += 0.00002
        exp_len = int(len(self.cpt)/self.card[rv])
        new_cpt = np.zeros(exp_len)

        rv_card = self.card[rv]
        rv_stride = self.stride[rv]

        k=0
        p = np.prod([self.card[r] for r in self.scope if self.stride[r] < self.stride[rv]])
        for i in range(exp_len):
            max_val = 0
            for c in range(rv_card):
                if self.cpt[k + (rv_stride*c)] > max_val:
                    max_val = self.cpt[k + (rv_stride*c)]
                    new_cpt[i] = max_val
            k+=1
            if (k % p == 0):
                k += (p* (rv_card - 1))
        self.cpt=new_cpt

        del self.card[rv]
        self.stride.update((k,v/rv_card) for k,v in self.stride.items() if v > rv_stride)
        del self.stride[rv]
        self.scope.remove(rv)

        #if rv == self.var:
            #self.var = [k for k,v in self.stride.items() if v==1][0]

        #if len(self.scope) > 0:
            #self.normalize()

    def reduce_factor_by_list(self, evidence):
        """
        Reduce the factor by numerous sets of
        [rv,val] -- this is done by running
        self.reduce_factor over the list of
        lists (*evidence*)

        Arguments
        ---------
        *evidence* : a list of lists/tuples
            The collection of rv-val pairs to
            remove from (condition upon) the factor


        Returns
        -------
        None

        Effects
        -------
        - see "self.reduce_factor"

        Notes
        -----
        - Again, might be good to check that each
            rv-val pair is actually in the factor
        """
        if isinstance(evidence, list):
            for rv,val in evidence:
                self.reduce_factor(rv,val)
        elif isinstance(evidence, dict):
            for rv,val in evidence.items():
                self.reduce_factor(rv,val)

    def reduce_factor(self, rv, val):
        """
        Condition the factor over evidence by eliminating any
        sets of values that don't align with [rv, val].

        This is different from "sumover_var" because "reduce_factor"
        is not summing over anything, it is simply removing any 
        parent-child instantiations which are not consistent with
        the evidence. Moreover, there should not be any need for
        normalization because the CPT should already be normalized
        over the rv-val evidence (but we do it anyways because of
        rounding)

        Note, this will completely eliminate "rv" from the factor,
        including from the scope and cpt.

        Arguments
        ---------
        *rv* : a string
            The random variable to eliminate/condition upon.

        *val* : a string
            The value of RV

        Returns
        -------
        None

        Effects
        -------
        - alters self.cpt
        - alters self.scope
        - alters self.card
        - alters self.stride

        Notes
        -----
        - There are no fail-safes here to make sure the
            rv-val pair is actually in the factor..

        """
        exp_len = len(self.cpt)/float(self.card[rv])
        new_cpt = np.zeros((exp_len,))

        val_idx = self.bn.F[rv]['values'].index(val)
        rv_card = self.card[rv]
        rv_stride = self.stride[rv]

        add_idx=0
        idx = val_idx*rv_stride
        while add_idx < exp_len:
            for j in self.cpt[idx:(idx+rv_stride)]:
                new_cpt[add_idx] = j
                add_idx+=1
            idx+=rv_card*rv_stride

        self.cpt=new_cpt
        del self.card[rv]
        self.stride.update((k,v/rv_card) for k,v in self.stride.items() if v > rv_stride)
        del self.stride[rv]
        self.scope.remove(rv)

        if rv == self.var:
            l = [k for k,v in self.stride.items() if v==1]
            if len(l)>0:
                self.var = l[0]

    def to_log(self):
        """
        Convert probabilities to log space from
        normal space.

        """
        self.cpt = np.round(np.log(self.cpt),5)

    def from_log(self):
        """
        Convert probabilities from log space to
        normal space.

        """
        self.cpt = np.round(np.exp(self.cpt),5)

    def perturb(self):
        """
        Add some noise to avoid "nan" when dividing by zero.
        This will probably make cpt values have many 
        decimal points (bad).
        """
        self.cpt += 1e-7

    def normalize(self):
        """
        Make relevant collections of probabilities sum to one.

        This function is ALWAYS going to normalize the variable
        for which the stride = 1, because it's assumed that's the
        main/child variable.

        Effects
        -------
        - alters self.cpt

        Notes
        -----

        """
        self.perturb() # stops nan's from happening
        var = [k for k,v in self.stride.items() if v==1]
        if len(var) > 0:
            var = var[0]
            for i in range(0,len(self.cpt),self.card[var]):
                temp_sum = float(np.sum(self.cpt[i:(i+self.card[var])]))
                for j in range(self.card[var]):
                    self.cpt[i+j] /= temp_sum
        else:
            for i in range(len(self.cpt)):
                self.cpt[i] /= (np.sum(self.cpt))







"""
************
Empirical
Distribution
Class
************

Create, maintain, and perform operations
over an empirical probability distribution
derived from a dataset. This class is most
useful for speeding up BN structure learning,
where distribution calculations can be cached
and therefore lookups become much faster over
repeated iterations.

References
----------
*** IMPORTANT ***
Daly and Shen,
"Methods to Accelerate the Learning of Bayesian Network Structures."
http://citeseerx.ist.psu.edu/viewdoc/download?
doi=10.1.1.127.7570&rep=rep1&type=pdf
*****************
"""

from __future__ import division

import numpy as np
from copy import copy, deepcopy


class EmpiricalDistribution(object):


	def __init__(self, data, names=None):

		if names is None:
			self.names = range(data.shape[1])
		else:
			assert (len(names) == self.NVAR), 'Passed-in names length must equal number of data columns'
			self.names = names

		self.NROW = data.shape[0]
		self.NVAR = data.shape[1]
		self.bins = [len(np.unique(data[:,n])) for n in range(self.NVAR)]
		

		hist,_ = np.histogramdd(data, bins=self.bins)
		self.counts = hist
		self.joint = (hist / hist.sum()) + 1e-3

		## COMPUTE MARGINAL FOR EACH VARIABLE ##
		#_range = range(self.NVAR)
		#for i,rv in enumerate(self.names):
		#	_axis = copy(_range)
		#	_axis.remove(i)
		#	self.marginal[rv] = np.sum(self.joint,axis=_axis)

		#self.marginal = dict([(rv, np.sum(self.joint,axis=i)) for i,rv in enumerate(self.names)])

		self.cache = {}

	def bayes_counts(self, bn):
		pass

	def idx_map(self, rvs):
		assert (isinstance(args, tuple)), "passed-in rvs must be a tuple"
		idx = [self.names.index(rv) for rv in rvs]
		return tuple(idx)

	def idx(self, rv):
		return self.names.index(rv)

	def axis_tuple(self, rv):
		pass

	def mpd(self, rv):
		"""
		Marginal Probability Distribtuion
		"""
		assert (isinstance(rv, str)), 'passed-in rv must be a string'
		if rv in self.cache:
			mpd = self.cache[rv]
		else:
			_axis = range(self.NVAR)
			_axis.remove(self.idx(rv))
			mpd = np.sum(self.joint, axis=tuple(_axis)) # CORRECT
			self.cache[rv] = mpd

		return mpd

	def jpd(self, rvs):
		assert (isinstance(args, tuple)), "passed-in rvs must be a tuple"
		if args in self.cache:
			jpd = self.cache[rvs]
		else:
			jpd = np.sum(self.joint, axis=self.idx_map(rvs)) # WRONG 
			self.cache[rvs] = jpd
			
		return jpd

	def cpd(self, lhs, rhs):
		assert (isinstance(rhs, tuple)), "passed-in rhs must be a tuple"

		_key = tuple((lhs,rhs))
		if _key in self.cache:
			cpd = self.cache[_key]
		else:
			try:
				tot = lhs + rhs
			except TypeError:
				tot = (lhs,) + rhs

			_numer = self.jpd(tot)
			_denom = self.jpd(rhs)
			cpd = _numer / (_denom + 1e-7)
			self.cache[_key] = cpd
		
		return cpd

	def mi(self, lhs, rhs, cond=None):
		"""
		Calculate mutual information with speedups
		by using caching for repeated lookups of
		joint/conditional distributions.

		"""
		bins = np.amax(data, axis=0) # read levels for each variable
		if len(bins) == 1:
			hist,_ = np.histogramdd(data, bins=(bins)) # frequency counts
			Px = hist/hist.sum()
			MI = -1 * np.sum( Px * np.log( Px ) )
			return round(MI, 4)
			
		if len(bins) == 2:
			hist,_ = np.histogramdd(data, bins=bins[0:2]) # frequency counts

			Pxy = hist / hist.sum()# joint probability distribution over X,Y,Z
			Px = np.sum(Pxy, axis = 1) # P(X,Z)
			Py = np.sum(Pxy, axis = 0) # P(Y,Z)	

			PxPy = np.outer(Px,Py)
			Pxy += 1e-7
			PxPy += 1e-7
			MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
			return round(MI,4)
		elif len(bins) > 2 and conditional==True:
			# CHECK FOR > 3 COLUMNS -> concatenate Z into one column
			if len(bins) > 3:
				data = data.astype('str')
				ncols = len(bins)
				for i in range(len(data)):
					data[i,2] = ''.join(data[i,2:ncols])
				data = data.astype('int')[:,0:3]

			bins = np.amax(data,axis=0)
			hist,_ = np.histogramdd(data, bins=bins) # frequency counts

			Pxyz = hist / hist.sum()# joint probability distribution over X,Y,Z
			Pz = np.sum(Pxyz, axis = (0,1)) # P(Z)
			Pxz = np.sum(Pxyz, axis = 1) # P(X,Z)
			Pyz = np.sum(Pxyz, axis = 0) # P(Y,Z)	

			Pxy_z = Pxyz / (Pz+1e-7) # P(X,Y | Z) = P(X,Y,Z) / P(Z)
			Px_z = Pxz / (Pz+1e-7) # P(X | Z) = P(X,Z) / P(Z)	
			Py_z = Pyz / (Pz+1e-7) # P(Y | Z) = P(Y,Z) / P(Z)

			Px_y_z = np.empty((Pxy_z.shape)) # P(X|Z)P(Y|Z)
			for i in range(bins[0]):
				for j in range(bins[1]):
					for k in range(bins[2]):
						Px_y_z[i][j][k] = Px_z[i][k]*Py_z[j][k]
			Pxyz += 1e-7
			Pxy_z += 1e-7
			Px_y_z += 1e-7
			MI = np.sum(Pxyz * np.log(Pxy_z / (Px_y_z)))
			
			return round(MI,4)
		elif len(bins) > 2 and conditional == False:
			data = data.astype('str')
			ncols = len(bins)
			for i in range(len(data)):
				data[i,1] = ''.join(data[i,1:ncols])
			data = data.astype('int')[:,0:2]

			hist,_ = np.histogramdd(data, bins=bins[0:2]) # frequency counts

			Pxy = hist / hist.sum()# joint probability distribution over X,Y,Z
			Px = np.sum(Pxy, axis = 1) # P(X,Z)
			Py = np.sum(Pxy, axis = 0) # P(Y,Z)	

			PxPy = np.outer(Px,Py)
			Pxy += 1e-7
			PxPy += 1e-7
			MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
			return round(MI,4)














"""
****************
BN Hybrid
Distance Metrics
****************

Code for computing the distance between 
2+ Bayesian networks that takes into account
both structure AND parameters. 

Any structure-based distance metric can be
combined with any parameter-based distance metric
to compute the hybrid distance.

Also, we support a type of supervised approach to
hybrid distance measurement where the relative importance
placed on the structure and distance metrics are varied
such that maximum (or minimum) closeness is achieved.

"""

def hybrid_distance(x,y, alpha, s='hamming', p='euclidean'):
	structure_score = hamming(x,y)
	parameter_score = euclidean(x,y)

	hybrid_score = structure_score*alpha + parameter_score*(1-alpha)

	return hybrid_score



"""
Lambda IAMB Code

References
----------
[1] Zhang et al, "An Improved IAMB Algorithm for Markov
Blanket Discovery"
"""

import numpy as np
from bamt.external.pyBN.utils.independence_tests import are_independent, entropy

def lambda_iamb(data, L=1.5, alpha=0.05, feature_selection=None):
	"""
	Lambda IAMB Algorithm for learning the structure of a
	Discrete Bayesian Network from data. This Algorithm
	is similar to the iamb algorithm, except that it allows
	for a "lambda" coefficient that helps avoid false positives.

	This algorithm was originally developed for use as a
	feature selection algorithm - discovering the markov
	blanket of a target variable is equivalent to discovering
	the relevant features for classifications.

	In practice, this algorithm does just as well as a feature
	selection method compared to IAMB when naive bayes was 
	used as a classifier, but Lambda-iamb actually does much
	better than traditional iamb when traditional iamb does
	very poorly due to high false positive rates.

	Arguments
	---------
	*data* : a nested numpy array

	*L* : a float
		The lambda hyperparameter - see [1].

	*alpha* : a float
		The type II error rate.

	Returns
	-------
	*bn* : a BayesNet object

	Effects
	-------
	None

	Notes
	-----
	"""
	n_rv = data.shape[1]
	Mb = dict([(rv,{}) for rv in range(n_rv)])

	if feature_selection is None:
		_T = range(n_rv)
	else:
		assert (not isinstance(feature_selection, list)), 'feature_selection must be only one value'
		_T = [feature_selection]

	# LEARN MARKOV BLANKET
	for T in _T:

		V = set(range(n_rv)) - {T}
		Mb_change=True

		# GROWING PHASE
		while Mb_change:
			Mb_change = False
			cols = tuple({T}) + tuple(Mb[T])
			H_tmb = entropy(data[:,cols])
			# find X1_min in V-Mb[T]-{T} that minimizes
			# entropy of T|X1_inMb[T]
			# i.e. min of entropy(data[:,(T,X,Mb[T])])
			min_val1, min_val2 = 1e7,1e7
			min_x1, min_x2 = None, None
			for X in V - Mb[T] - {T}:
				cols = (T,X)+tuple(Mb[T])
				ent_val = entropy(data[:,cols])
				if ent_val < min_val:
					min_val2, min_val1 = min_val1, ent_val
					min_x2,min_x1 = min_x1, X
					
			# if min_x1 is dependent on T given Mb[T]...
			cols = (min_x1,T) + tuple(Mb[T])
			if are_independent(data[:,cols]):
				if (min_val2 - L*min_val1) < ((1-L)*H_tmb):
					cols = (min_x2,T)+tuple(Mb[T])
					if are_independent(data[:,cols]):
						Mb[T].add(min_x1)
						Mb[T].add(min_x2)
						Mb_change = True
			else:
				Mb[T].add(X)
				Mb_change = True

		# SHRINKING PHASE
		for X in Mb[T]:
			# if x is indepdent of t given Mb[T] - {x}
			cols = (X,T) + tuple(Mb[T]-{X})
			if mi_test(data[:,cols]) > alpha:
				Mb[T].remove(X)

	if feature_selection is None:
		# RESOLVE GRAPH STRUCTURE
		edge_dict = resolve_markov_blanket(Mb, data)

		# ORIENT EDGES
		oriented_edge_dict = orient_edges_Mb(edge_dict,Mb,data,alpha)

		# CREATE BAYESNET OBJECT
		value_dict = dict(zip(range(data.shape[1]),
			[list(np.unique(col)) for col in data.T]))
		bn=BayesNet(oriented_edge_dict,value_dict)

		return bn
	else:
		return Mb[_T]











import configparser
from os import path, stat

config = configparser.ConfigParser()

CONFIGFILE = path.join(path.dirname(path.abspath(__file__)), 'selbst.ini')

if path.isfile(CONFIGFILE) and stat(CONFIGFILE).st_size != 0:
    config.read(CONFIGFILE)
else:
    open(CONFIGFILE, 'a').close()
    # config['NODES'] = {'models_storage': path.join(path.dirname(path.abspath(__file__)), 'Nodes_data')}
    config['NODES'] = {'models_storage': path.join(path.expanduser("~"), 'BAMT', 'Nodes_data')}
    config['LOG'] = {"log_conf_loc": path.join(path.dirname(path.abspath(__file__)), 'logging.conf')}
    with open(CONFIGFILE, 'w') as configfile:
        config.write(configfile)




__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.classes.factor import Factor
from bamt.external.pyBN.classes.factorization import Factorization
from bamt.external.pyBN.utils.graph import *

from copy import deepcopy, copy
import numpy as np
import json


def exact_bp(bn, target=None, evidence=None, downward_pass=False):
	"""
	Perform Belief Propagation (Message Passing) over a Clique Tree. This
	is sometimes referred to as the "Junction Tree Algorithm" or
	the "Hugin Algorithm".

	It involves an Upward Pass (see [1] pg. 353) along with
	Downward Pass (Calibration) ([1] pg. 357) if the target involves
	multiple random variables - i.e. is a list

	Steps Involved:
		1. Build a Clique Tree from a Bayesian Network
			a. Moralize the BN
			b. Triangulate the graph
			c. Find maximal cliques and collapse into nodes
			d. Create complete graph and make edge weights = sepset cardinality
			e. Using Max Spanning Tree to create a tree of cliques
		2. Assign each factor to only one clique
		3. Compute the initial potentials of each clique
			- multiply all of the clique's factors together
		4. Perform belief propagation based on message passing protocol.


	Arguments
	---------
	*bn* : a BayesNet object

	Returns
	-------


	Notes
	-----

	"""
	# 1: Moralize the graph
	# 2: Triangluate
	# 3: Build a clique tree using max spanning
	# 4: Propagation of probabilities using message passing

	# creates clique tree and assigns factors, thus satisfying steps 1-3
	ctree = CliqueTree(bn) # might not be initialized?
	#G = ctree.G
	#cliques = copy.copy(ctree.V)

	# select a clique as root where target is in scope of root
	root = ctree.V[0]
	if target is not None:
		for v in ctree.V:
			if target in ctree[v].scope:
				root = v
				break

	
	clique_ordering = ctree.dfs_postorder(root=root)

	# UPWARD PASS
	# send messages up the tree from the leaves to the single root
	for i in clique_ordering:
		#clique = ctree[i]
		for j in ctree.parents(i):
			ctree[i] >> ctree[j] 
			#clique.send_message(ctree[j])
		# if root node, collect its beliefs
		#if len(ctree.parents(i)) == 0:
			#ctree[root].collect_beliefs()
	ctree[root].collect_beliefs()
	marginal_target = ctree[root].marginalize_over(target)

	# DOWNWARD PASS
	if downward_pass == True:
		# send messages down the tree from the root to the leaves
		# (not needed unless *target* involves more than one variable)
		new_ordering = list(reversed(clique_ordering))
		for j in new_ordering:
			for i in ctree.children(j):
				ctree[j] >> ctree[i]
			# if leaf node, collect its beliefs
			if len(ctree.children(j)) == 0:                    
				ctree[j].collect_beliefs()

	return marginal_target

	# beliefs hold the answers


"""
*****************************
Maximum Likelihood Estimation
Parameter Learning
*****************************

"""
from __future__ import division

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np

def mle_fast(bn, data, nodes=None, counts=False, np=False):
	"""
	Maximum Likelihood estimation that is about 100 times as
	fast as the original mle_estimator function - but returns
	the same result
	"""
	def merge_cols(data, cols):
		if len(cols) == 1:
			return data[cols[0]]
		elif len(cols) > 1:
			data = data[cols].astype('str')
			ncols = len(cols)
			for i in range(len(data)):
				data.ix[i,0] = ''.join(data.ix[i,0:ncols])
			data = data.astype('int')
			return data.ix[:,0]

	if nodes is None:
		nodes = list(bn.nodes())
	else:
		if not isinstance(nodes, list):
			nodes = list(nodes)

	F = dict([(rv, {}) for rv in nodes])
	for i, n in enumerate(nodes):
		F[n]['values'] = list(np.unique(data.ix[:,i]))
		bn.F[n]['values'] = list(np.unique(data.ix[:,i]))

	for rv in nodes:
		parents = bn.parents(rv)
		if len(parents)==0:
			if np:
				F[rv]['cpt'] = np.histogram(data.ix[:,rv], bins=bn.card(rv))[0]
			else:
				F[rv]['cpt'] = list(np.histogram(data.ix[:,rv], bins=bn.card(rv))[0])
		else:
			if np:
				F[rv]['cpt'] = np.histogram2d(merge_cols(data,parents),data.ix[:,rv], 
				bins=[np.prod([bn.card(p) for p in parents]),bn.card(rv)])[0].flatten()
			else:
				F[rv]['cpt'] = list(np.histogram2d(merge_cols(data,parents),data.ix[:,rv], 
					bins=[np.prod([bn.card(p) for p in parents]),bn.card(rv)])[0].flatten())
	if counts:
		return F
	else:
		for rv in nodes:
			F[rv]['parents'] = [var for var in nodes if rv in bn.E[var]]
			for i in range(0,len(F[rv]['cpt']),bn.card(rv)):
				temp_sum = float(np.sum(F[rv]['cpt'][i:(i+bn.card(rv))]))
				for j in range(bn.card(rv)):
					F[rv]['cpt'][i+j] /= (temp_sum+1e-7)
					F[rv]['cpt'][i+j] = round(F[rv]['cpt'][i+j],5)
		bn.F = F




def mle_estimator(bn, data, nodes=None, counts=False):
	"""
	Maximum Likelihood Estimation is a frequentist
	method for parameter learning, where there is NO
	prior distribution. Instead, the frequencies/counts
	for each parameter start at 0 and are simply incremented
	as the relevant parent-child values are observed in the
	data. 

	This can be a risky method for small datasets, because if a 
	certain parent-child instantiation is never observed in the
	data, then its probability parameter will be ZERO (even if you
	know it should at least have a very small probability). 

	Note that the Bayesian and MLE estimators essentially converge
	to the same set of values as the size of the dataset increases.

	Also note that, unlike the structure learning algorithms, the
	parameter learning functions REQUIRE a passed-in BayesNet object
	because there MUST be some pre-determined structure for which
	we can actually learn the parameters. You can't learn parameters
	without structure - so structure must always be there first!

	Finally, note that this function can be used to calculate only
	ONE conditional probability table in a BayesNet object by
	passing in a subset of random variables with the "nodes"
	argument - this is mostly used for score-based structure learning,
	where a single cpt needs to be quickly recalculate after the
	addition/deletion/reversal of an arc.

	Arguments
	---------
	*bn* : a BayesNet object
		The associated network structure for which
		the parameters will be learned

	*data* : a nested numpy array

	*nodes* : a list of strings
		Which nodes to learn the parameters for - if None,
		all nodes will be used as expected.

	Returns
	-------
	None

	Effects
	-------
	- modifies/sets bn.data to the learned parameters

	Notes
	-----
	- Currently doesn't return correct solution

	data attributes:
		"numoutcomes" : an integer
	    "vals" : a list
	    "parents" : a list or None
	    "children": a list or None
	    "cprob" : a nested python list

	- Do not want to alter bn.data directly!

	"""
	if nodes is None:
		nodes = list(sorted(bn.nodes()))
	else:
		if not isinstance(nodes, list):
			nodes = list(nodes)

	F = dict([(rv, {}) for rv in nodes])
	for i, n in enumerate(nodes):
		F[n]['values'] = list(np.unique(data[:,i]))
		bn.F[n]['values'] = list(np.unique(data[:,i]))

	obs_dict = dict([(rv,[]) for rv in nodes])
	# set empty conditional probability table for each RV
	for rv in nodes:
		# get number of values in the CPT = product of scope vars' cardinalities
		p_idx = int(np.prod([bn.card(p) for p in bn.parents(rv)])*bn.card(rv))
		F[rv]['cpt'] = [0]*p_idx
		bn.F[rv]['cpt'] = [0]*p_idx
	
	# loop through each row of data
	for row in data:
		# store the observation of each variable in the row
		for rv in nodes:
			obs_dict[rv] = row[rv]
		
		#obs_dict = dict([(rv,row[rv]) for rv in nodes])
		# loop through each RV and increment its observed parent-self value
		for rv in nodes:
			rv_dict= { n: obs_dict[n] for n in obs_dict if n in bn.scope(rv) }
			offset = bn.cpt_indices(target=rv,val_dict=rv_dict)[0]
			F[rv]['cpt'][offset]+=1

	if counts:
		return F
	else:
		for rv in nodes:
			F[rv]['parents'] = [var for var in nodes if rv in bn.E[var]]
			for i in range(0,len(F[rv]['cpt']),bn.card(rv)):
				temp_sum = float(np.sum(F[rv]['cpt'][i:(i+bn.card(rv))]))
				for j in range(bn.card(rv)):
					F[rv]['cpt'][i+j] /= (temp_sum+1e-7)
					F[rv]['cpt'][i+j] = round(F[rv]['cpt'][i+j],5)
		bn.F = F









__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.classes.factor import Factor
from bamt.external.pyBN.classes.factorization import Factorization
from bamt.external.pyBN.utils.graph import *

from copy import deepcopy, copy
import numpy as np
import json

def marginal_ve_e(bn, target, evidence={}):
	"""
	Perform Sum-Product Variable Elimination on
	a Discrete Bayesian Network.

	Arguments
	---------
	*bn* : a BayesNet object

	*target* : a list of target RVs

	*evidence* : a dictionary, where
		key = rv and value = rv value

	Returns
	-------
	*marginal_dict* : a dictionary, where
		key = an rv in target and value =
		a numpy array containing the key's
		marginal conditional probability distribution.

	Notes
	-----
	- Mutliple pieces of evidence often returns "nan"...numbers too small?
		- dividing by zero -> perturb values in Factor class
	"""
	_phi = Factorization(bn)

	order = copy(list(bn.nodes()))
	order.remove(target)

	#### EVIDENCE PROCESSING ####
	for E, e in evidence.items():
		_phi -= (E,e)
		order.remove(E)

	#### SUM-PRODUCT ELIMINATE VAR ####
	for var in order:
		_phi /= var

	# multiply phi's together if there is evidence
	final_phi = _phi.consolidate()

	return np.round(final_phi.cpt,4)


# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module contains tools for representing "crazy" nodes -- nodes where the method for sampling is to multiply the crazyinput by -10 or 10 and add :math:`\pi` -- as class instances with their own *choose* method to choose an outcome for themselves based on parent outcomes. 

The existence of this 'crazy' type is meant to indicate the true universality of
the universal sampling method found in :doc:`hybayesiannetwork`. While no CPD would
actually be this crazy, the libary has the setup to support any type of CPD.


'''
import math
import random

class Crazy():
    '''
    This class represents a crazy node, as described above. It contains the *Vdataentry* attribute and the *choose* method.

    '''
    def __init__(self, Vdataentry):
        '''
        This class is constructed with the argument *Vdataentry* which must be a dict containing a dictionary entry for this particualr node. The dict must contain an entry of the following form::

            "crazyinput": <number that is the crazy input>

        This ``"crazyinput"`` entry contains the number that will be used in the crazy sampling function.  The *Vdataentry* attribute is set equal to this *Vdataentry* input upon instantiation.
        '''
        self.Vdataentry = Vdataentry
        '''A dict containing CPD data for the node.'''
        
    def choose(self, pvalues):
        '''
        Randomly choose state of node from probability distribution conditioned on *pvalues*.

        This method has two parts: (1) determining the proper probability
        distribution, and (2) using that probability distribution to determine
        an outcome.

        Arguments:
            1. *pvalues* -- An array containing the assigned states of the node's parents. This must be in the same order as the parents appear in self.Vdataentry['parents'].

        The function takes the crazyinput, multiplies it by either 10 or -10 randomly, adds :math:`\\pi`, converts it to a string, and appends the word "bluberries!". It returns this value.

        '''
        crazyinput = self.Vdataentry["crazyinput"]
        answer = "%.2f blueberries!" % (random.choice([10, -10]) * crazyinput + math.pi)
        return answer



"""
*************
Read BayesNet
from File
*************

Read a BayesNet object from a file. These
functions all currently work, but support
for more formats should be added in the future.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartouth.edu>"""

import json

from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.utils.graph import topsort


def read_bn(path):
    """
    Wrapper function for reading BayesNet objects
    from various file types.

    Arguments
    ---------
    *path* : a string
        The path (relative or absolute) - MUST
        include extension -> that's how we know
        which file reader to call.

    Returns
    -------
    *bn* : a BayesNet object

    Effects
    -------
    None

    Notes
    -----

    """
    if '.bif' in path:
        return read_bif(path)
    elif '.bn' in path:
        return read_json(path)
    elif '.mat' in path:
        return read_mat(path)
    else:
        print("Path Extension not recognized")

def read_bif(path):
    """
    This function reads a .bif file into a
    BayesNet object. It's probably not the 
    fastest or prettiest but it gets the job
    done.

    Arguments
    ---------
    *path* : a string
        The path

    Returns
    -------
    *bn* : a BayesNet object

    Effects
    -------
    None

    Notes
    -----
    *V* : a list of strings
    *E* : a dict, where key = vertex, val = list of its children
    *F* : a dict, where key = rv, val = another dict with
                keys = 'parents', 'values', cpt'

    """
    _parents = {} # key = vertex, value = list of vertices in the scope (includind itself)
    _cpt = {} # key = vertex, value = list (then numpy array)
    _vals = {} # key=vertex, val=list of its possible values

    with open(path, 'r') as f:
        while True:
            line = f.readline()
            if 'variable' in line:
                new_vertex = line.split()[1]

                _parents[new_vertex] = []
                _cpt[new_vertex] = []
                #_vals[new_vertex] = []

                new_line = f.readline()
                new_vals = new_line.replace(',', ' ').split()[6:-1] # list of vals
                _vals[new_vertex] = new_vals
                num_outcomes = len(new_vals)
            elif 'probability' in line:
                line = line.replace(',', ' ')
                child_rv = line.split()[2]
                parent_rvs = line.split()[4:-2]

                if len(parent_rvs) == 0: # prior
                    new_line = f.readline().replace(';', ' ').replace(',',' ').split()
                    prob_values = new_line[1:]
                    _cpt[child_rv].append(map(float,prob_values))
                    #_cpt[child_rv] = map(float,prob_values)
                else: # not a prior
                    _parents[child_rv].extend(list(parent_rvs))
                    while True:
                        new_line = f.readline()
                        if '}' in new_line:
                            break
                        new_line = new_line.replace(',',' ').replace(';',' ').replace('(', ' ').replace(')', ' ').split()
                        prob_values = new_line[-(len(_vals[new_vertex])):]
                        prob_values = map(float,prob_values)
                        _cpt[child_rv].append(prob_values)
            if line == '':
                break

    # CREATE FACTORS
    _F = {}
    _E = {}
    for rv in _vals.keys():
        _E[rv] = [c for c in _vals.keys() if rv in _parents[c]]
        f = {
            'parents' : _parents[rv],
            'values' : _vals[rv],
            'cpt' : [item for sublist in _cpt[rv] for item in sublist]
        }
        _F[rv] = f

    bn = BayesNet()
    bn.F = _F
    bn.E = _E
    bn.V = list(topsort(_E))

    return bn

def read_json(path):
    """
    Read a BayesNet object from the json format. This
    format has the ".bn" extension and is completely
    unique to pyBN.

    Arguments
    ---------
    *path* : a string
        The file path

    Returns
    -------
    None

    Effects
    -------
    - Instantiates and sets a new BayesNet object

    Notes
    -----
    
    This function reads in a libpgm-style format into a bn object

    File Format:
        {
            "V": ["Letter", "Grade", "Intelligence", "SAT", "Difficulty"],
            "E": [["Intelligence", "Grade"],
                ["Difficulty", "Grade"],
                ["Intelligence", "SAT"],
                ["Grade", "Letter"]],
            "Vdata": {
                "Letter": {
                    "ord": 4,
                    "numoutcomes": 2,
                    "vals": ["weak", "strong"],
                    "parents": ["Grade"],
                    "children": None,
                    "cprob": [[.1, .9],[.4, .6],[.99, .01]]
                },
                ...
        }


    """
    def byteify(input):
        if isinstance(input, dict):
            return {byteify(key):byteify(value) for key,value in input.iteritems()}
        elif isinstance(input, list):
            return [byteify(element) for element in input]
        elif isinstance(input, unicode):
            return input.encode('utf-8')
        else:
            return input

    bn = BayesNet()
    
    f = open(path,'r')
    ftxt = f.read()

    success=False
    try:
        data = byteify(json.loads(ftxt))
        bn.V = data['V']
        bn.E = data['E']
        bn.F = data['F']
        success = True
    except ValueError:
        print("Could not read file - check format")
    bn.V = topsort(bn.E)

    return bn

def read_mat(path, delim=' '):
    """
    Read an adjacency matrix into a BayesNet object.

    NOTE: This is for reading the structure only, and
    therefore no parameters for the BayesNet object will
    be set - they must be learned by calling "mle_estimator"
    or "bayes_estimator" on the object.
    """
    _V = []
    _E = {}
    _F = {}
    with open(path, 'r') as f:
        for line in f:
            line = line.split(delim)
            rv = line[0]
            _E[rv] = []

    bn = BayesNet(_E)

    return bn




















"""
******************************
Conditional Independence Tests 
for Constraint-based Learning
******************************

Implemented Constraint-based Tests
----------------------------------
- mutual information
- Pearson's X^2

I may consider putting this code into its own class structure. The
main benefit I could see from doing this would be the ability to 
cache joint/marginal/conditional probabilities for expedited tests.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np
from scipy import stats

from bamt.external.pyBN.utils.data import unique_bins


def are_independent(data, alpha=0.05, method='mi_test'):
    pval = mi_test(data)
    if pval < alpha:
        return True
    else:
        return False


def mutual_information(data, conditional=False):
    # bins = np.amax(data, axis=0)+1 # read levels for each variable
    bins = unique_bins(data)
    if len(bins) == 1:
        hist, _ = np.histogramdd(data, bins=(bins))  # frequency counts
        Px = (hist / hist.sum()) + (1e-7)
        MI = -1 * np.sum(Px * np.log(Px))
        return round(MI, 4)

    if len(bins) == 2:
        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Px = np.sum(Pxy, axis=1)  # P(X,Z)
        Py = np.sum(Pxy, axis=0)  # P(Y,Z)

        PxPy = np.outer(Px, Py)
        Pxy += 1e-7
        PxPy += 1e-7
        MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
        return round(MI, 4)
    elif len(bins) > 2 and conditional == True:
        # CHECK FOR > 3 COLUMNS -> concatenate Z into one column
        if len(bins) > 3:
            data = data.astype('str')
            ncols = len(bins)
            for i in range(len(data)):
                data[i, 2] = ''.join(data[i, 2:ncols])
            data = data.astype(np.int64)[:, 0:3]

        bins = np.amax(data, axis=0)
        hist, _ = np.histogramdd(data, bins=bins)  # frequency counts

        Pxyz = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Pz = np.sum(Pxyz, axis=(0, 1))  # P(Z)
        Pxz = np.sum(Pxyz, axis=1)  # P(X,Z)
        Pyz = np.sum(Pxyz, axis=0)  # P(Y,Z)

        Pxy_z = Pxyz / (Pz + 1e-7)  # P(X,Y | Z) = P(X,Y,Z) / P(Z)
        Px_z = Pxz / (Pz + 1e-7)  # P(X | Z) = P(X,Z) / P(Z)
        Py_z = Pyz / (Pz + 1e-7)  # P(Y | Z) = P(Y,Z) / P(Z)

        Px_y_z = np.empty((Pxy_z.shape))  # P(X|Z)P(Y|Z)
        for i in range(bins[0]):
            for j in range(bins[1]):
                for k in range(bins[2]):
                    Px_y_z[i][j][k] = Px_z[i][k] * Py_z[j][k]
        Pxyz += 1e-7
        Pxy_z += 1e-7
        Px_y_z += 1e-7
        MI = np.sum(Pxyz * np.log(Pxy_z / (Px_y_z)))

        return round(MI, 4)
    elif len(bins) > 2 and conditional == False:
        data = data.astype('str')
        ncols = len(bins)
        for i in range(len(data)):
            data[i, 1] = ''.join(data[i, 1:ncols])
        data = data.astype(np.int64)[:, 0:2]

        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Px = np.sum(Pxy, axis=1)  # P(X,Z)
        Py = np.sum(Pxy, axis=0)  # P(Y,Z)

        PxPy = np.outer(Px, Py)
        Pxy += 1e-7
        PxPy += 1e-7
        MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
        return round(MI, 4)


def mi_test(data, test=True):
    """
    This function performs the mutual information (cross entropy)-based
    CONDITIONAL independence test. Because it is conditional, it requires
    at LEAST three columns. For the marginal independence test, use
    "mi_test_marginal".

    We use the maximum likelihood estimators as probabilities. The
    mutual information value is computed, then the
    chi square test is used, with degrees of freedom equal to
    (|X|-1)*(|Y|-1)*Pi_z\inZ(|z|).

    This function works on datasets that contain MORE than three
    columns by concatenating the extra columns into one. For that
    reason, it is a little slower in that case.

    For two variables only:

    This function performs mutual information (cross entropy)-based
    MARGINAL independence test. Because it is marginal, it requires
    EXACTLY TWO columns. For the conditional independence test, use
    "mi_test_conditional".

    This is the same as calculated the KL Divergence, i.e.
    I(X,Y) = Sigma p(x,y)* log p(x,y) *( p(x)/p(y) )

    NOTE: pval < 0.05 means DEPENDENCE, pval > 0.05 means INDEPENDENCE.
    In other words, the pval represent the probability this relationship
    could have happened at random or by chance. if the pval is very small,
    it means the two variables are likely dependent on one another.

    Steps:
        - Calculate the marginal/conditional probabilities
        - Compute the Mutual Information value
        - Calculate chi2 statistic = 2*N*MI
        - Compute the degrees of freedom
        - Compute the chi square p-value

    Arguments
    ----------
    *data* : a nested numpy array
        The data from which to learn - must have at least three
        variables. All conditioned variables (i.e. Z) are compressed
        into one variable.

    Returns
    -------
    *p_val* : a float
        The pvalue from the chi2 and ddof

    Effects
    -------
    None

    Notes
    -----
    - Doesn't currently work with strings...
    - Should generalize to let data be a Pandas DataFrame --> would
    encourage external use.

    """

    # bins = np.amax(data, axis=0)+1 # read levels for each variable
    bins = unique_bins(data)
    if len(bins) == 2:
        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        # Pxy = hist / hist.sum()# joint probability distribution over X,Y,Z
        Pxy = hist / data.shape[0]
        Px = np.sum(Pxy, axis=1)  # P(X,Z)
        Py = np.sum(Pxy, axis=0)  # P(Y,Z)

        PxPy = np.outer(Px, Py)
        Pxy += 1e-7
        PxPy += 1e-7
        MI = np.sum(Pxy * np.log(Pxy / (PxPy)))
        if not test:
            return round(MI, 4)
        else:
            chi2_statistic = 2 * len(data) * MI
            ddof = (bins[0] - 1) * (bins[1] - 1)
            p_val = 2 * stats.chi2.pdf(chi2_statistic, ddof)
            return round(p_val, 4)
    else:
        # CHECK FOR > 3 COLUMNS -> concatenate Z into one column
        if len(bins) > 3:
            data = data.astype('str')
            ncols = len(bins)
            for i in range(len(data)):
                data[i, 2] = ''.join(data[i, 2:ncols])
            data = data.astype(np.int64)[:, 0:3]

        # bins = np.amax(data,axis=0)
        bins = unique_bins(data)
        hist, _ = np.histogramdd(data, bins=bins)  # frequency counts

        # Pxyz = hist / hist.sum()# joint probability distribution over X,Y,Z
        Pxyz = hist / data.shape[0]
        Pz = np.sum(Pxyz, axis=(0, 1))  # P(Z)
        Pxz = np.sum(Pxyz, axis=1)  # P(X,Z)
        Pyz = np.sum(Pxyz, axis=0)  # P(Y,Z)

        Pxy_z = Pxyz / (Pz + 1e-7)  # P(X,Y | Z) = P(X,Y,Z) / P(Z)
        Px_z = Pxz / (Pz + 1e-7)  # P(X | Z) = P(X,Z) / P(Z)
        Py_z = Pyz / (Pz + 1e-7)  # P(Y | Z) = P(Y,Z) / P(Z)

        Px_y_z = np.empty((Pxy_z.shape))  # P(X|Z)P(Y|Z)
        for i in range(bins[0]):
            for j in range(bins[1]):
                for k in range(bins[2]):
                    Px_y_z[i][j][k] = Px_z[i][k] * Py_z[j][k]
        Pxyz += 1e-7
        Pxy_z += 1e-7
        Px_y_z += 1e-7
        MI = np.sum(Pxyz * np.log(Pxy_z / (Px_y_z)))
        if not test:
            return round(MI, 4)
        else:
            chi2_statistic = 2 * len(data) * MI
            ddof = (bins[0] - 1) * (bins[1] - 1) * bins[2]
            p_val = 2 * stats.chi2.pdf(chi2_statistic, ddof)  # 2* for one tail
            return round(p_val, 4)


def entropy(data):
    """
    In the context of structure learning, and more specifically
    in constraint-based algorithms which rely on the mutual information
    test for conditional independence, it has been proven that the variable
    X in a set which MAXIMIZES mutual information is also the variable which
    MINIMIZES entropy. This fact can be used to reduce the computational
    requirements of tests based on the following relationship:

        Entropy is related to marginal mutual information as follows:
            MI(X;Y) = H(X) - H(X|Y)

        Entropy is related to conditional mutual information as follows:
            MI(X;Y|Z) = H(X|Z) - H(X|Y,Z)

        For one varibale, H(X) is equal to the following:
            -1 * sum of p(x) * log(p(x))

        For two variables H(X|Y) is equal to the following:
            sum over x,y of p(x,y)*log(p(y)/p(x,y))

        For three variables, H(X|Y,Z) is equal to the following:
            -1 * sum of p(x,y,z) * log(p(x|y,z)),
                where p(x|y,z) = p(x,y,z)/p(y)*p(z)
    Arguments
    ----------
    *data* : a nested numpy array
        The data from which to learn - must have at least three
        variables. All conditioned variables (i.e. Z) are compressed
        into one variable.

    Returns
    -------
    *H* : entropy value

    """
    try:
        cols = data.shape[1]
    except IndexError:
        cols = 1

    bins = np.amax(data, axis=0)
    if isinstance(bins, np.ndarray):
        for i in range(len(bins)):
            if bins[i] == 0:
                bins[i] = 1
    else:
        bins = 1
    # bins = unique_bins(data)

    if cols == 1:
        hist, _ = np.histogramdd(data, bins=(bins))  # frequency counts
        Px = hist / hist.sum() + (1e-7)
        H = -1 * np.sum(Px * np.log(Px))

    elif cols == 2:  # two variables -> assume X then Y
        hist, _ = np.histogramdd(data, bins=bins[0:2])  # frequency counts

        Pxy = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Py = np.sum(Pxy, axis=0)  # P(Y)
        Py += 1e-7
        Pxy += 1e-7
        H = np.sum(Pxy * np.log(Py / Pxy))

    else:
        # CHECK FOR > 3 COLUMNS -> concatenate Z into one column
        if cols > 3:
            data = data.astype('str')
            ncols = len(bins)
            for i in range(len(data)):
                data[i, 2] = ''.join(data[i, 2:ncols])
            data = data.astype(np.int64)[:, 0:3]

        bins = np.amax(data, axis=0)
        hist, _ = np.histogramdd(data, bins=bins)  # frequency counts

        Pxyz = hist / hist.sum()  # joint probability distribution over X,Y,Z
        Pyz = np.sum(Pxyz, axis=0)

        Pxyz += 1e-7  # for log -inf
        Pyz += 1e-7
        H = -1 * np.sum(Pxyz * np.log(Pxyz)) + np.sum(Pyz * np.log(Pyz))

    return round(H, 4)


def mi_from_en(data):
    """
    Calculate Mutual Information based on entropy alone. This
    function isn't really faster than calculating MI from
    "mi_test" due to overhead in the histogram/binning from
    numpy, but it's mostly here for entropy validation or
    if "mi_test" breaks. BUT, calling "entropy" once is much
    faster than calling "mi_test" once, so that is where the
    speedup occurs.

    This has been validated with both 2, 3 and 4 variables.

    Entropy is related to marginal mutual information as follows:
            MI(X;Y) = H(X) - H(X|Y)
    Entropy is related to conditional mutual information as follows:
            MI(X;Y|Z) = H(X|Z) - H(X|Y,Z)
    """
    ncols = data.shape[1]

    if ncols == 1:
        print("Need at least 2 columns")

    elif ncols == 2:
        MI = entropy(data[:, 0]) - entropy(data)

    elif ncols == 3:
        MI = entropy(data[:, (0, 2)]) - entropy(data)

    elif ncols > 3:
        # join extra columns
        data = data.astype('str')
        ncols = data.shape[1]
        for i in range(len(data)):
            data[i, 2] = ''.join(data[i, 2:ncols])
        data = data.astype(np.int64)[:, 0:3]

        MI = entropy(data[:, (0, 2)]) - entropy(data)

    return round(MI, 4)


def chi2_test(data):
    """
    Test null hypothesis that P(X,Y,Z) = P(Z)P(X|Z)P(Y|Z)
    versus empirically observed P(X,Y,Z) in the data using
    the traditional chisquare test based on observed versus
    expected frequency bins.

    Steps
        - Calculate P(XYZ) empirically and expected
        - Compute ddof
        - Perfom one-way chisquare

    Arguments
    ---------
    *data* : a nested numpy array
        The data from which to learn - must have at least three
        variables. All conditioned variables (i.e. Z) are compressed
        into one variable.

    Returns
    -------
    *chi2_statistic* : a float
        Chisquare statistic
    *p_val* : a float
        The pvalue from the chi2 and ddof

    Effects
    -------
    None

    Notes
    -----
    - Assuming for now that |Z| = 1... generalize later
    - Should generalize to let data be a Pandas DataFrame --> would
    encourage external use.

    """
    # compress extra Z variables at the start.. not implemented yet
    # bins = np.amax(data, axis=0)+1
    bins = unique_bins(data)
    hist, _ = np.histogramdd(data, bins=bins)

    Pxyz = hist / hist.sum()  # joint probability distribution over X,Y,Z

    Pz = np.sum(Pxyz, axis=(0, 1))  # P(Z)
    Pxz = np.sum(Pxyz, axis=1)  # P(X,Z)
    Pyz = np.sum(Pxyz, axis=0)  # P(Y,Z)

    Px_z = Pxz / (Pz + 1e-7)  # P(X | Z) = P(X,Z) / P(Z)
    Py_z = Pyz / (Pz + 1e-7)  # P(Y | Z) = P(Y,Z) / P(Z)

    observed_dist = Pxyz  # Empirical distribution
    # Not correct right now -> Pz is wrong dimension
    Px_y_z = np.empty((Pxy_z.shape))  # P(Z)P(X|Z)P(Y|Z)
    for i in range(bins[0]):
        for j in range(bins[1]):
            for k in range(bins[2]):
                Px_y_z[i][j][k] = Px_z[i][k] * Py_z[j][k]
    Px_y_z *= Pz

    observed = observed_dist.flatten() * len(data)
    expected = expected_dist.flatten() * len(data)

    ddof = (bins[0] - 1) * (bins[1] - 1) * bins[2]
    chi2_statistic, p_val = stats.chisquare(observed, expected, ddof=ddof)

    return chi2_statistic, p_val



"""
These are functions for dealing with datasets
that have strings as values - as those values
must be converted to integers in order to be 
used in many structure learning functions, for
example.

There is, of course, the issue of how to get those
string values back into the underlying representation
after the learning occurs..
"""
import numpy as np
from copy import copy

def replace_strings(data, return_values=False):
	"""
	Takes a dataset of strings and returns the
	integer mapping of the string values.

	It might be useful to return a dictionary of the
	string values so that they can be mapped back in
	the actual BayesNet object when necessary.
	"""
	data = copy(data)
	i = 0
	value_dict = {}
	for col in data.T:
		vals,_,ret, = np.unique(col, True, True)
		data[:,i] = ret
		value_dict[i] = list(vals)
		i+=1

	if return_values:
		return np.array(data,dtype=np.int32), value_dict
	else:
		return np.array(data, dtype=np.int32)

def unique_bins(data):
	"""
	Get the unique values for each column in a dataset.
	"""
	bins = np.empty(len(data.T), dtype=np.int32)
	i = 0
	for col in data.T:
		bins[i] = len(np.unique(col))
		i+=1
	return bins


import os,sys,inspect
# currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# parentdir = os.path.dirname(currentdir)
# sys.path.insert(0,parentdir)


from copy import copy
import math
from typing import List
import numpy as np
import pandas as pd
from bamt.external.pyBN.utils.independence_tests import mutual_information, entropy
from bamt.preprocess.discretization import get_nodes_type
from bamt.preprocess.numpy_pandas import loc_to_DataFrame
from bamt.preprocess.graph import edges_to_dict


def query_filter(data: pd.DataFrame, columns: List, values: List):
    """
    Filters the data according to the column-value list
            Arguments
    ----------
    *data* : pandas.DataFrame
    Returns
    -------
    *data_trim* : pandas.DataFrame
        Filtered data.
    Effects
    -------
    None
    """
    data_copy = copy(data)
    filter_str = '`' + str(columns[0]) + '`' + ' == ' + str(values[0])
    if len(columns) == 1:
        return data_copy.query(filter_str)
    else:
        for i in range(1, len(columns)):
            filter_str += ' & ' + '`' + str(columns[i]) + '`' + ' == ' + str(values[i])
        data_trim = data_copy.query(filter_str)
        return data_trim


def entropy_gauss(pd_data):
    """
    Calculate entropy for Gaussian multivariate distributions.
            Arguments
    ----------
    *data* : pd.DataFrame
    Returns
    -------
    *entropy* : a float
        The entropy for Gaussian multivariate distributions.
    Effects
    -------
    None
    """
    if not isinstance(pd_data, pd.Series):
        data = copy(pd_data).values.T 
    else:
        data = np.array(copy(pd_data)).T
    if data.size == 0: 
        return 0.0
    flag_row = False
    flag_col = False
    
    if isinstance(data[0], np.float64):
        flag_row = True
    elif (len(data[0]) < 2) | (data.ndim < 2) :
        flag_row = True
    elif data.shape[0] < 2:
        
        flag_row = True
    if isinstance(copy(data).T[0], np.float64):
        flag_col = True
    elif (len(copy(data).T) < 2) | (copy(data).T.ndim < 2):
        flag_col = True
    elif data.shape[1] < 2:
        
        flag_col = True

    if flag_row & flag_col:
        return sys.float_info.max
    elif flag_row | flag_col:
        var = np.var(data)
        if var > 1e-16:
            return 0.5 * (1 + math.log(var*2*math.pi))
        else:
            return sys.float_info.min
    else:
        var = np.linalg.det(np.cov(data))  
        N = var.ndim      
        if var > 1e-16:
            return 0.5 * (N * (1 + math.log(2*math.pi)) + math.log(var))
        else:
            return sys.float_info.min


def entropy_all(data, method = 'MI'):
    """
        For one varibale, H(X) is equal to the following:
            -1 * sum of p(x) * log(p(x))
        For two variables H(X|Y) is equal to the following:
            sum over x,y of p(x,y)*log(p(y)/p(x,y))
        For three variables, H(X|Y,Z) is equal to the following:
            -1 * sum of p(x,y,z) * log(p(x|y,z)),
                where p(x|y,z) = p(x,y,z)/p(y)*p(z)
    Arguments
    ----------
    *data* : pd.DataFrame
    Returns
    -------
    *H* : entropy value"""
    if type(data) is np.ndarray:
        return entropy_all(loc_to_DataFrame(data), method = method)
    elif isinstance(data, pd.Series):
        return entropy_all(pd.DataFrame(data), method)
    elif (type(data) is pd.DataFrame):
        nodes_type = get_nodes_type(data)
        column_disc = []
        for key in nodes_type:
            if nodes_type[key] == 'disc':
                column_disc.append(key)
        column_cont = []
        for key in nodes_type:
            if nodes_type[key] == 'cont':
                column_cont.append(key)
        data_disc = data[column_disc]
        data_cont = data[column_cont]
        
        if len(column_cont) == 0:
                return entropy(data_disc.values)
        elif len(column_disc) == 0:
            return entropy_gauss(data_cont)
        else:      
            H_disc = entropy(data_disc.values)
            dict_comb = {}
            comb_prob = {}
            for i in range(len(data_disc)):
                row = data_disc.iloc[i]
                comb = ''
                for _, val in row.items():
                    comb = comb + str(val) + ', '
                if not comb in dict_comb:
                    dict_comb[comb] = row
                    comb_prob[comb] = 1
                else:
                    comb_prob[comb] += 1
            
            H_cond = 0.0
            for key in list(dict_comb.keys()):
                filtered_data = query_filter(data, column_disc, list(dict_comb[key]))
                filtered_data = filtered_data[column_cont]
                if comb_prob[key] == 1:
                    if (method == 'BIC') | (method == 'AIC'):
                        H_cond += comb_prob[key] / len(data_disc) * entropy_gauss(data[column_cont])
                    else:
                        H_cond += comb_prob[key] / len(data_disc) * sys.float_info.max
                else:
                    H_cond += comb_prob[key] /len(data_disc) * entropy_gauss(filtered_data)
                if (method == 'BIC') | (method == 'AIC'):
                    if H_cond > entropy_gauss(data[column_cont]):
                        H_cond = entropy_gauss(data[column_cont])
                    
            return (H_disc + H_cond)

def entropy_cond(data, column_cont, column_disc, method):
    data_cont = data[column_cont]
    data_disc = data[column_disc]
    H_gauss = entropy_gauss(data_cont)
    H_cond = 0.0
                
    dict_comb = {}
    comb_prob = {}
    for i in range(len(data_disc)):
        row = data_disc.iloc[i]
        comb = ''
        for _, val in row.items():
            comb = comb + str(val) + ', '
        if not comb in dict_comb:
            dict_comb[comb] = row
            comb_prob[comb] = 1
        else:
            comb_prob[comb] += 1
    
    for key in list(dict_comb.keys()):
        filtered_data = query_filter(data, column_disc, list(dict_comb[key]))
        filtered_data = filtered_data[column_cont]
        if comb_prob[key] == 1:
            if (method == 'BIC') | (method == 'AIC'):
                H_cond += comb_prob[key] / len(data_disc) * H_gauss
            else:
                H_cond += comb_prob[key] / len(data_disc) * sys.float_info.max
        else:
            H_cond += comb_prob[key] / len(data_disc) * entropy_gauss(filtered_data)
    if (method == 'BIC') | (method == 'AIC'):
        if H_cond > H_gauss:
            return H_gauss
        else:
            return H_cond
    return H_cond


def mi_gauss(data, method='MI', conditional=False):
    """
    Calculate Mutual Information based on entropy. 
    In the case of continuous uses entropy for Gaussian multivariate distributions.
            Arguments
    ----------
    *data* : pandas.DataFrame
    Returns
    -------
    *MI* : a float
        The Mutual Information
    Effects
    -------
    None
    Notes
    -----
    - Need to preprocess data with code_categories
    """
    if type(data) is np.ndarray:
        return mi_gauss(loc_to_DataFrame(data), method, conditional)
    elif isinstance(data, pd.Series):
        return(mi_gauss(pd.DataFrame(data)))
    elif type(data) is pd.DataFrame:
        nodes_type = get_nodes_type(data)
        if conditional:
            #Hill-Climbing does not use conditional MI, but other algorithms may require it
            #At the moment it counts on condition of the last row in the list of columns
            print('Warning: conditional == True')
            nodes_type_trim = copy(nodes_type)
            data_trim = copy(data)
            list_keys = list(nodes_type_trim.keys)
            del nodes_type_trim[list_keys[-1]]
            del data_trim[list_keys[-1]]
            return (mi_gauss(data, nodes_type, method) - mi_gauss(data_trim, nodes_type, method))
        else:
            column_disc = []
            for key in nodes_type:
                if nodes_type[key] == 'disc':
                    column_disc.append(key)
            column_cont = []
            for key in nodes_type:
                if nodes_type[key] == 'cont':
                    column_cont.append(key)
            data_disc = data[column_disc]
            data_cont = data[column_cont]

            H_gauss = 0.0
            H_cond = 0.0
            
            if len(column_cont) == 0:
                return(mutual_information(data_disc.values, conditional = False))
            elif len(column_disc) == 0:
                if len(column_cont) == 1:
                    return entropy_gauss(data_cont)
                else:
                    data_last = data_cont[[column_cont[-1]]]
                    column_cont_trim = copy(column_cont)
                    del column_cont_trim[-1]
                    data_cont_trim = data[column_cont_trim]
                
                    H_gauss = entropy_gauss(data_last) + entropy_gauss(data_cont_trim)-entropy_gauss(data_cont)
                    H_gauss = min(H_gauss, entropy_gauss(data_last), entropy_gauss(data_cont_trim))
                    #H_gauss = entropy_gauss(data_cont)
                    H_cond = 0.0
            else:
                H_gauss = entropy_gauss(data_cont)
                H_cond = entropy_cond(data, column_cont, column_disc, method)
                
            return(H_gauss-H_cond)

def mi(edges: list, data: pd.DataFrame, method='MI'):
    """
    Bypasses all nodes and summarizes scores, 
    taking into account the parent-child relationship.
            Arguments
    ----------
    *edges* : list
    *data* : pd.DataFrame
    Returns
    -------
    *sum_score* : float
    Effects
    -------
    None
    """
    parents_dict = edges_to_dict(edges)
    sum_score = 0.0
    nodes_with_edges = parents_dict.keys()
    for var in nodes_with_edges:
        child_parents = [var]
        child_parents.extend(parents_dict[var])
        sum_score += mi_gauss(copy(data[child_parents]), method)
    nodes_without_edges = list(set(data.columns).difference(set(nodes_with_edges)))
    for var in nodes_without_edges:
        sum_score += mi_gauss(copy(data[var]), method)
    return sum_score


"""
***********
Naive Bayes
***********

Learn both the structure AND parameters
of a Naive Bayes Model from data

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np
from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.learning.parameter.mle import mle_estimator
from bamt.external.pyBN.learning.parameter.bayes import bayes_estimator


def naive_bayes(data, target, estimator='mle'):
	"""
	Learn naive bayes model from data.

	The Naive Bayes model is a Tree-based
	model where all random variables have
	the same parent (the "target" variable).
	From a probabilistic standpoint, the implication
	of this model is that all random variables 
	(i.e. features) are assumed to be
	conditionally independent of any other random variable,
	conditioned upon the single parent (target) variable.

	It turns out that this model performs quite well
	as a classifier, and can be used as such. Moreover,
	this model is quite fast and simple to learn/create
	from a computational standpoint.

	Note that this function not only learns the structure,
	but ALSO learns the parameters.

	Arguments
	---------
	*data* : a nested numpy array

	*target* : an integer
		The target variable column in *data*

	Returns
	-------
	*bn* : a BayesNet object,
		with the structure instantiated.

	Effects
	-------
	None

	Notes
	-----

	"""	
	value_dict = dict(zip(range(data.shape[1]),
		[list(np.unique(col)) for col in data.T]))

	edge_dict = {target:[v for v in value_dict if v!=target]}
	edge_dict.update(dict([(rv,[]) for rv in value_dict if rv!=target]))

	bn = BayesNet(edge_dict,value_dict)
	if estimator == 'bayes':
		bayes_estimator(bn,data)
	else:
		mle_estimator(bn,data)
	return bn




















"""
Collection of Graph Algorithms.

Any networkx dependencies should exist here only, but
there are currently a few exceptions which will
eventually be corrected.

"""

import networkx as nx
import numpy as np
from copy import copy

def would_cause_cycle(e, u, v, reverse=False):
	"""
	Test if adding the edge u -> v to the BayesNet
	object would create a DIRECTED (i.e. illegal) cycle.
	"""
	G = nx.DiGraph(e)
	if reverse:
		G.remove_edge(v,u)
	G.add_edge(u,v)
	try:
		nx.find_cycle(G, source=u)
		return True
	except:
		return False



def topsort(edge_dict, root=None):
	"""
	List of nodes in topological sort order from edge dict
	where key = rv and value = list of rv's children
	"""
	queue = []
	if root is not None:
		queue = [root]
	else:
		for rv in edge_dict.keys():
			prior=True
			for p in edge_dict.keys():
				if rv in edge_dict[p]:
					prior=False
			if prior==True:
				queue.append(rv)
	
	visited = []
	while queue:
		vertex = queue.pop(0)
		if vertex not in visited:
			visited.append(vertex)
			for nbr in edge_dict[vertex]:
				queue.append(nbr)
			#queue.extend(edge_dict[vertex]) # add all vertex's children
	return visited

def dfs_postorder(edge_dict, root=None):
	return list(reversed(topsort(edge_dict, root)))

def mst(edge_dict):
	"""
	Calcuate Minimum Spanning Tree
	for a weighted edge dictionary,
	where key = rv, value = another dictionary
	where key = child of rv, value = edge weight
	between rv -> child edge.

	Arguments
	---------
	*w_edge_dict* : a dictionary, where
		key = rv, value = another dictionary, where
		key = child node of "rv", value = integer/float
		edge weight between the "rv" -> "child" edge

	Returns
	-------
	*mst_edge_dict* : a non-weighted edge dict
		The minimum spanning tree from w_edge_dict, stored
		as a dictionary where key = rv, value = list of
		rv's children

	Effects
	-------
	None

	Notes
	-----
	test:
		input:
			g = {0:{1:3,3:2,8:4},1:{7:4},
				2:{3:6,7:2,5:1},3:{4:1},4:{8:8},
				5:{6:8},6:{},7:{2:2},8:{}}
		output:
			g = {0:[3,1,8],1:[7],2:[5],3:[4],4:[],
				5:[6],6:[],7:[2],8:[]}
		
	 
	"""
	mst_G = dict([(rv,[]) for rv in range(len(edge_dict))])
	nrv = len(mst_G)
	reached = [0]
	unreached = range(1,nrv)

	for k in range(nrv):
		source, sink = None, None
		min_cost = np.inf
		for rn in reached:
			e = edge_dict[rn].items()
			e.sort(key = lambda x : x[1])
			for sn, weight in e:
				if sn in unreached and weight < min_cost:
					min_cost = weight
					source = rn
					sink = sn
					break

		if source is None or sink is None:
			break

		# Add e to the minimum spanning tree
		mst_G[source].append(sink)
		#if undirected == True:
		#	mst_G[sink].append(source)

		# Mark newly include node as reached
		unreached.remove(sink)
		reached.append(sink)
	
	return mst_G

def make_chordal(bn, v=None,e=None):
	"""
	This function creates a chordal graph - i.e. one in which there
	are no cycles with more than three nodes.

	Can supply a v_list and e_list for chordal graph of any random graph..

	We start from the moral graph, so if that is already chordal then it
	will return that.

	Algorithm from Cano & Moral 1990 ->
	'Heuristic Algorithms for the Triangulation of Graphs'


	Parameters
	----------
	*v* : a list (optional)
	    A vertex list

	*e* : a list (optional)
	    An edge list


	Returns
	-------
	*G* : a network Digraph object

	Effects
	-------
	None

	Notes
	-----
	Where is this used? Do we need to use networkx?       


	"""
	chordal_E = bn.moralized_edges() # start with moral graph

	# if moral graph is already chordal, no need to alter it
	if not is_chordal(chordal_E):            
		temp_E = copy(chordal_E)
		temp_V = []

		# if v and e is supplied, skip all the rest
		if v and e:
			chordal_E = copy(e)
			temp_E = copy(chordal_E)
			temp_V = copy(v)
		else:
			temp_G = nx.Graph()
			temp_G.add_edges_from(chordal_E)
			degree_dict = temp_G.degree()
			temp_V = sorted(degree_dict, key=degree_dict.get)
		#print temp_V
		for v in temp_V:
			#Add links between the pairs nodes adjacent to Node i
			#Add those links to chordal_E and temp_E
			adj_v = set([n for e in temp_E for n in e if v in e and n!=v])
			for a1 in adj_v:
				for a2 in adj_v:
					if a1!=a2:
						if [a1,a2] not in chordal_E and [a2,a1] not in chordal_E:
							chordal_E.append([a1,a2])
							temp_E.append([a1,a2])
			# remove Node i from temp_V and all its links from temp_E 
			temp_E2 = []
			for edge in temp_E:
				if v not in edge:
					temp_E2.append(edge)
			temp_E = temp_E2

	
	G = nx.Graph()
	G.add_edges_from(chordal_E)
	return G


def is_chordal(edge_list):
	"""
	Check if the graph is chordal/triangulated.

	Parameters
	----------
	*edge_list* : a list of lists (optional)
	The edges to check (if not self.E)

	Returns
	-------
	*nx.is_chordal(G)* : a boolean
	Whether the graph is chordal or not

	Effects
	-------
	None

	Notes
	-----
	Again, do we need networkx for this? Eventually should
	write this check on our own.

	"""
	G = nx.Graph(list(edge_list))
	return nx.is_chordal(G)


"""
********
UnitTest
OrientEdges
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from external.pyBN.structure_learn.orient_edges import orient_edges_gs, orient_edges_pc


class OrientEdgesTestCase(unittest.TestCase):

	def setUp(self):
		pass
	def tearDown(self):
		pass

	def test_orient_edges_pc(self):
		e = {0:[1,2],1:[0],2:[0]}
		b = {0: [], 1: {2: (0,)}, 2: {1: (0,)}}
		self.assertDictEqual(orient_edges_pc(e,b),
			{0: [1, 2], 1: [], 2: []})

	def test_orient_edges_gs(self):
		e = {0: [1, 2], 1: [0], 2: [0]}
		b = {0: [1, 2], 1: [0], 2: [0]}
		dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		path = (os.path.join(dpath,'lizards.csv'))
		data = np.loadtxt(path, dtype='int32',skiprows=1,delimiter=',')
		self.assertDictEqual(orient_edges_gs(e,b,data,0.05),
			{0: [1, 2], 1: [], 2: []})


"""
********
UnitTest
Writing
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest


def suite():
	suite = unittest.TestLoader().loadTestsFromTestCase(WritingTestCase)
	return suite


class WritingTestCase(unittest.TestCase):

	def setUp(self):
		pass
	def tearDown(self):
		pass


# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools to represent and handle Bayesian networks with conditional probability distributions that can be specified node-by-node. 

This method allows for the construction of a Bayesian network with every combination of every type of CPD, provided that the user provides a method for sampling each type of node and stores this method in the proper place, namely as the ``choose()`` method of a class in ``libpgm.CPDtypes/``.

'''

import random

from bamt.external.libpgm.orderedskeleton import OrderedSkeleton


class HyBayesianNetwork(OrderedSkeleton):
    '''
    This class represents a Bayesian network with CPDs of any type. The nodes of the Bayesian network are stored first in a dictionary, specifying their "type", which should be descriptive ('discrete', 'lg', etc.). Furthermore, the types of each node associate them with a class found in ``libpgm/CPDtypes/``. The nodes are then stored also as instances of classes found in this directory. The purpose of this is that each node has its own method for being sampled given the outcomes of its parents.
    
    '''

    def __init__(self, orderedskeleton=None, nodedata=None):
        '''
        This class can be called either with or without arguments. If it is called without arguments, none of its attributes are instantiated and it is left to the user to instantiate them manually. If it is called with arguments, the attributes will be loaded directly from the inputs. The arguments must be (in order):

            1. *orderedskeleton* -- An instance of the :doc:`OrderedSkeleton <orderedskeleton>` or :doc:`GraphSkeleton <graphskeleton>` (as long as it's ordered) class.
            2. *nodedata* -- An instance of the :doc:`NodeData <nodedata>` class.
        
        It is required that the *nodedata* class instance inputted has its *nodes* attribute instantiated. In order for this to be the case, the instance must have run its *entriestoinstances* method.

        If the arguments above are present, all attributes of the class (*V*, *E*, *Vdata*, and *nodes*) will be automatically copied from the graph skeleton and node data inputs.

        Upon loading, the class will also check that the keys of *Vdata* correspond to the vertices in *V*.

        '''
        if (orderedskeleton != None and nodedata != None):
            try:
                self.V = orderedskeleton.V
                '''A list of the names of the vertices.'''
                self.E = orderedskeleton.E
                '''A list of [origin, destination] pairs of vertices that make edges.'''
                self.Vdata = nodedata.Vdata
                '''A dictionary containing CPD data for the nodes.'''

                # specific to hybrid Bayesian network
                self.nodes = nodedata.nodes
                '''A dictionary of {key: value} pairs linking the node name (the key) to a class instance (the value) representing the node, its node data, and its sampling function.'''
            except:
                raise (Exception,
                       "Inputs were malformed; first arg must contain V and E attributes and second arg must contain Vdata and nodes attributes.")

            # check that inputs match up
            assert sorted(self.V) == sorted(self.Vdata.keys()), "Node data did not match graph skeleton nodes."

    def randomsample(self, n, method, evidence=None):
        '''
        Produce *n* random samples from the Bayesian networki, subject to *evidence*, and return them in a list. This function requires the *nodes* attribute to be instantiated.
        
        This function takes the following arguments:

            1. *n* -- The number of random samples to produce.
            2. *evidence* -- (Optional) A dict containing (vertex: value) pairs that describe the evidence. To be used carefully because it does manually overrides the nodes with evidence instead of affecting the joint probability distribution of the entire graph.
        
        And returns:
            A list of *n* independent random samples, each element of which is a dict containing (vertex: value) pairs.
        
        Usage example: this would generate a sequence of 10 random samples::
            
            import json

            from external.libpgm.nodedata import NodeData
            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.hybayesiannetwork import HyBayesianNetwork

            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittesthdict.txt")   # an input file
            skel.load("../tests/unittestdict.txt")

            # topologically order graphskeleton
            skel.toporder()

            # convert nodes to class instances
            nd.entriestoinstances()

            # load bayesian network
            hybn = HyBayesianNetwork(skel, nd)

            # sample 
            result = hybn.randomsample(10)

            # output
            print json.dumps(result, indent=2)
            


        '''
        assert ((isinstance(n, int) and n > 0), "Argument must be a positive integer.")

        random.seed()
        seq = []
        for _ in range(n):
            outcome = dict()
            for vertex in self.V:
                outcome[vertex] = "default"

            def assignnode(name, node):

                # check if node is already observed
                if (evidence != None):
                    if name in evidence.keys():
                        return evidence[name]

                # get parent values
                p = self.getparents(name)
                if (p == []):
                    pvalues = []
                else:
                    pvalues = [outcome[t] for t in self.Vdata[name]["parents"]]  # ideally can we pull this from the skeleton so as not to store parent data at all?
                    for pvalue in pvalues:
                        assert (pvalue != 'default', "Graph skeleton was not topologically ordered.")

                # use built in function to determine outcome
                
                return node.choose(pvalues, method, outcome)

            for s in self.V:
                if (outcome[s] == "default"):
                    outcome[s] = assignnode(s, self.nodes[s])

            seq.append(outcome)
        return seq



# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools for collecting and managing sets of samples generated by the library's sampling functions. By averaging a series of samples, the progam can approximate a joint probability distribution without having to do the exact calculations, which may be useful in large networks. 

'''


class SampleAggregator(object):
    '''
    This class is a machine for aggregating data from sample sequences. It contains the method *aggregate*.
    
    '''
    def __init__(self):
        self.seq = None
        '''The sequence inputted.'''
        self.avg = None
        '''The average of all the entries in *seq*, represented as a dict where each vertex has an entry whose value is a dict of {key, value} pairs, where each key is a possible outcome of that vertex and its value is the approximate frequency.'''


    def aggregate(self, samplerstatement):
        '''
        Generate a sequence of samples using *samplerstatement* and return the average of its results. 
        
        Arguments:
            1. *samplerstatement* -- The statement of a function (with inputs) that would output a sequence of samples. For example: ``bn.randomsample(50)`` where ``bn`` is an instance of the :doc:`DiscreteBayesianNetwork <discretebayesiannetwork>` class.
        
        This function stores the output of *samplerstatement* in the attribute *seq*, and then averages *seq* and stores the approximate distribution found in the attribute *avg*. It then returns *avg*. 
       
        Usage example: this would print the average of 10 data points::

            import json

            from external.libpgm.nodedata import NodeData
            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.discretebayesiannetwork import DiscreteBayesianNetwork
            from external.libpgm.sampleaggregator import SampleAggregator

            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittestdict.txt")
            skel.load("../tests/unittestdict.txt")

            # topologically order graphskeleton
            skel.toporder()

            # load bayesian network
            bn = DiscreteBayesianNetwork(skel, nd)

            # build aggregator
            agg = SampleAggregator()

            # average samples
            result = agg.aggregate(bn.randomsample(10))

            # output
            print json.dumps(result, indent=2)

        '''
        
        # get sequence
        seq = samplerstatement
        
        # denominator
        denom = len(seq)
        
        output = dict()
        for key in list(seq[0].keys()):
            output[key] = dict()
            for trial in seq:
                keyss = list(output[key].keys())
                vall = trial[key]
                if (keyss.count(vall) > 0):
                    output[key][trial[key]] += 1
                else:
                    output[key][trial[key]] = 1
                    
            # normalize
            for entry in list(output[key].keys()):
                output[key][entry] = output[key][entry] / float(denom)
        
        self.seq = seq
        self.avg = output
        
        return output



# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools to represent and handle Bayesian networks with linear Gaussian conditional probability distributions.

A linear Gaussian distribution means that the node has a continuous range of outcomes, with a normal distribution over those outcomes. This normal distribution can be parameterized by a *mean* and a *variance*. A linear Gaussian means that the variance of the node is fixed, while the mean is a linear function of the outcomes of the node's parents. In math terms, the mean :math:`m(u)` of a node *u* is a linear function of the values :math:`x_1,\\dots,x_n` of its parents, each weighted by some coefficient :math:`\\beta_i`:

.. math::

    m(u) = \\beta_0 +\\beta_1 x_1 + \\dots + \\beta_n x_n

Linear Gaussians are simple but widely used in statistical modeling.

'''

import random
import math
import sys

from bamt.external.libpgm.orderedskeleton import OrderedSkeleton


class LGBayesianNetwork(OrderedSkeleton):
    '''
    This class represents a Bayesian network with linear Gaussian CPDs. It contains the attributes *V*, *E*, and *Vdata*, as well as the method *randomsample*. 
   
    '''

    def __init__(self, orderedskeleton=None, nodedata=None):
        '''
        This class can be called either with or without arguments. If it is called without arguments, none of its attributes are instantiated and it is left to the user to instantiate them manually. If it is called with arguments, the attributes will be loaded directly from the inputs. The arguments must be (in order):

            1. *orderedskeleton* -- An instance of the :doc:`OrderedSkeleton <orderedskeleton>` or :doc:`GraphSkeleton <graphskeleton>` (as long as it's ordered) class.
            2. *nodedata* -- An instance of the :doc:`NodeData <nodedata>` class.

        If these arguments are present, all attributes of the class (*V*, *E*, and *Vdata*) will be automatically copied from the graph skeleton and node data inputs.
        
        This class requires that the *Vdata* attribute gets loaded with a dictionary with node data of the following fomat::

            "vertex": {
                "parents": ["<name of parent 1>", ... , "<name of parent n>"],
                "children": ["<name of child 1>", ... , "<name of child n>"],   
                "mean_base": <the base mean of the Gaussian distribution>,
                "mean_scal": [<scalar for parent 1 outcome>, ... , <scalar for parent n outcome>],
                "variance": <variance of the Gaussian distibution>
            }

        Note that additional keys are possible in the dict of each vertex.

        Upon loading, the class will also check that the keys of *Vdata* correspond to the vertices in *V*.

        '''
        if (orderedskeleton != None and nodedata != None):
            try:
                self.V = orderedskeleton.V
                '''A list of the names of the vertices.'''
                self.E = orderedskeleton.E
                '''A list of [origin, destination] pairs of vertices that make edges.'''
                self.Vdata = nodedata.Vdata
                '''A dictionary containing CPD data for the nodes.'''
            except: 
                raise (Exception, "Inputs were malformed; first arg must contain V and E attributes and second arg must contain Vdata attribute.")

            assert sorted(self.V) == sorted(self.Vdata.keys()), "Vertices did not match node data"

    def randomsample(self, n, evidence=None, mode="normal"):
        '''
        Produce *n* random samples from the Bayesian Network and return them in a list. 
       
        See above for how the means of linear Gaussians are calculated during sampling.

        This function takes the following arguments:

            1. *n* -- The number of random samples to produce.
            2. *evidence* -- (Optional) A dict containing (vertex: value) pairs that describe the evidence. To be used carefully because it does manually overrides the nodes with evidence instead of affecting the joint probability distribution of the entire graph.
            3. *mode* -- (Optional) Can be set to "verbose", whereupon the method will return a [value, mean, variance] list for each node rather than just the actual value.  
        
        And returns:
            A list of *n* independent random samples, each element of which is a dict containing (vertex: value) pairs.

        Usage example: this would generate a sequence of 10 random samples::
            
            import json

            from external.libpgm.nodedata import NodeData
            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.lgbayesiannetwork import LGBayesianNetwork

            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittestlgdict.txt")  # an input file
            skel.load("../tests/unittestdict.txt")

            # topologically order graphskeleton
            skel.toporder()

            # load bayesian network
            lgbn = LGBayesianNetwork(skel, nd)

            # sample 
            result = lgbn.randomsample(10)

            # output
            print json.dumps(result, indent=2)

        '''
        assert (isinstance(n, int) and n > 0), "Argument must be a positive integer."

        random.seed()
        seq = []
        distribseq = []
        for _ in range(n):
            outcome = dict()
            distribs = dict()
            for vertex in self.V:
                outcome[vertex] = "default"

            def assignnode(s):
            
                if (evidence != None):
                    if s in evidence.keys():
                        return [evidence[s], ["given value", "given value"]]

                # calculate Bayesian parameters (mean and variance)
                mean = self.Vdata[s]["mean_base"]
                if (self.Vdata[s]["parents"] != None):
                    for x in range(len(self.Vdata[s]["parents"])):
                        parent = self.Vdata[s]["parents"][x]
                        assert outcome[parent] != 'default', "Graph skeleton was not topologically ordered."
                        mean += outcome[parent] * self.Vdata[s]["mean_scal"][x]
                variance = self.Vdata[s]["variance"]

                distribution = [mean, variance]

                # draw random outcome from Gaussian 
                return [random.gauss(mean, math.sqrt(variance)), distribution]          

            for s in self.V:
                if (outcome[s] == "default"):
                    pair = assignnode(s)
                    outcome[s] = pair[0]
                    distribs[s] = pair[1]
            seq.append(outcome)
            distribseq.append(distribs)

        if mode == "normal":
            return seq

        elif mode == "verbose":
            result = []
            for x in range(len(seq)):
                result.append(dict())
                for node in seq[x]: 
                     result[x][node] = [seq[x][node], distribseq[x][node][0], distribseq[x][node][1]] 
            return result



from bamt.external.pyBN.classes import *
from bamt.external.pyBN.classification import *
from bamt.external.pyBN.inference import *
from bamt.external.pyBN.io import *
from bamt.external.pyBN.learning import *
#from external.pyBN.plotting import *
from bamt.external.pyBN.utils import *


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bamt',
 'bamt.external',
 'bamt.external.libpgm',
 'bamt.external.pyBN',
 'bamt.external.pyBN.classes',
 'bamt.external.pyBN.classes._tests',
 'bamt.external.pyBN.classification',
 'bamt.external.pyBN.inference',
 'bamt.external.pyBN.inference._tests',
 'bamt.external.pyBN.inference.map_exact',
 'bamt.external.pyBN.inference.marginal_approx',
 'bamt.external.pyBN.inference.marginal_exact',
 'bamt.external.pyBN.io',
 'bamt.external.pyBN.io._tests',
 'bamt.external.pyBN.learning',
 'bamt.external.pyBN.learning.parameter',
 'bamt.external.pyBN.learning.parameter._tests',
 'bamt.external.pyBN.learning.structure',
 'bamt.external.pyBN.learning.structure._tests',
 'bamt.external.pyBN.learning.structure.constraint',
 'bamt.external.pyBN.learning.structure.exact',
 'bamt.external.pyBN.learning.structure.hybrid',
 'bamt.external.pyBN.learning.structure.naive',
 'bamt.external.pyBN.learning.structure.score',
 'bamt.external.pyBN.learning.structure.tree',
 'bamt.external.pyBN.plotting',
 'bamt.external.pyBN.utils',
 'bamt.external.pyBN.utils._tests',
 'bamt.preprocess',
 'bamt.utils']

package_data = \
{'': ['*']}

install_requires = \
['gmr==1.6.2',
 'matplotlib==3.5.1',
 'missingno>=0.5.1,<0.6.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas==1.4.0',
 'pgmpy>=0.1.17,<0.2.0',
 'pomegranate==0.14.8',
 'pyitlib>=0.2.2,<0.3.0',
 'pyvis>=0.1.9,<0.2.0',
 'scikit-learn==1.0.2',
 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'bamt',
    'version': '1.1.24',
    'description': 'data modeling and analysis tool based on Bayesian networks',
    'long_description': "# BAMT\nRepository of a data modeling and analysis tool based on Bayesian networks\n\n\nBAMT - Bayesian Analytical and Modelling Toolkit. This repository contains a data modeling and analysis tool based on Bayesian networks. It can be divided into two main parts - algorithms for constructing and training Bayesian networks on data and algorithms for applying Bayesian networks for filling gaps, generating synthetic data, and searching for anomalous values.\n\n### Installation\n\nBAMT package is available via PyPi: ``pip install bamt``\n\nBayesian network learning\n=========================\nIn terms of training Bayesian networks on data, the following algorithms are implemented:\n- Building the structure of a Bayesian network based on expert knowledge by directly specifying the structure of the network;\n- Building the structure of a Bayesian network on data using three algorithms - Hill Climbing, evolutionary and PC. For Hill Climbing, the following score functions are implemented - MI, K2, BIC, AIC. The algorithms work on both discrete and mixed data.\n- Training the parameters of distributions in the nodes of the network on the basis of data.\n\nDifference from existing implementations:\n- Algorithms work on mixed data;\n- Structural learning implements score-functions for mixed data;\n- Parametric learning implements the use of a mixture of Gaussian distributions to approximate continuous distributions;\n- The algorithm for structural training of large Bayesian networks (> 10 nodes) is based on local training of small networks with their subsequent algorithmic connection.\n![title](img/BN_gif.gif)\n\nGenerating synthetic data\n=========================\nIn terms of data analysis and modeling using Bayesian networks, a pipeline has been implemented to generate synthetic data by sampling from Bayesian networks.\n![title](img/synth_gen.png)\n\nOil and Gas Reservoirs Parameters Analysis\n==========================================\nBayesian networks can be used to restore gaps in reservoirs data, search for anomalous values, and also to search for analogous reservoirs.\n![title](img/concept.png)\n\n# Project structure\n## Utils\n  1. GraphUtils consist of functions for:\n  - Finding nodes types/signs\n  - Topological ordering\n  2. MathUtils consist of functions for:\n  - Additional function to support calculation of metrics from group1 ('MI', 'LL', 'BIC', 'AIC')\n\n## Preprocessing\nPreprocessor module allows user to transform data according pipeline (in analogy with pipeline in scikit-learn).\n## Networks\nThere are 3 general types of networks: discrete, gaussian, hybrid. All implemented in networks.py  \n## Nodes\nContains nodes' classes and their methods.\n",
    'author': 'Roman Netrogolov',
    'author_email': 'romius2001@gmail.com',
    'maintainer': 'N Mramorov',
    'maintainer_email': None,
    'url': 'https://github.com/ITMO-NSS-team/BAMT',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)




__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np


def loopy_bp(target=None, evidence=None, max_iter=100):
	"""
	Perform Message Passing (Loopy Belief Propagation) 
	over a cluster graph.

	This is a good candidate for Numba JIT.

	See Koller pg. 397.


	Parameters
	----------


	Returns
	-------


	Notes
	-----
	- Copied from clustergraph class... not tested.
	- Definitely a problem due to normalization (prob vals way too small)
	- Need to check the scope w.r.t. messages.. all clusters should not
	be accumulating rv's in their scope over the course of the algorithm.
	"""
	def collect_beliefs(cgraph):
		cgraph.beliefs = {}
		for cluster in self.V:
			cgraph.V[cluster].collect_beliefs()
			#print('Belief ' , cluster , ' : \n', self.V[cluster].belief.cpt)
			cgraph.beliefs[cluster] = cgraph.V[cluster].belief

	# 1: Moralize the graph
	# 2: Triangluate
	# 3: Build a clique tree using max spanning
	# 4: Propagation of probabilities using message passing

	# creates clique tree and assigns factors, thus satisfying steps 1-3
	cgraph = ClusterGraph(bn)

	edge_visit_dict = dict([(i,0) for i in cgraph.E])

	iteration = 0
	while not cgraph.is_calibrated():
		if iteration == max_iter:
			break
		if iteration % 50 == 0:
			print('Iteration: ' , iteration)
			for cluster in cgraph.V.values():
				cluster.collect_beliefs()
		# select an edge
		e_idx = np.random.randint(0,len(cgraph.E))
		edge_select = cgraph.E[e_idx]
		p_idx = np.random.randint(0,2)
		parent_edge = edge_select[p_idx]
		child_edge = edge_select[np.abs(p_idx-1)]
		print(parent_edge , child_edge)

		# send a message along that edge
		cgraph.V[parent_edge].send_message(cgraph.V[child_edge])

		iteration += 1
	print('Now Collecting Beliefs..')
	collect_beliefs(cgraph)
	#bn.ctree = self


"""
Various metrics for evaluating
score-based search for Bayesian
Networks structure learning.

Generally, score-based structure learning
involves finding the structure which maximizes
some function of the likelihood of the data, given
the structure and parameters of the learned BN.

It is important to take advantage of the decomposability
of the scores. That is, if you add/delete/reverse an edge,
then you only need to re-calculate the score based on that
local change (usually just the child node's CPT) to get the
difference from the original graph. 

Here are a few which are (or can be) implemented:

Information-theoretic scoring functions:
	LL (Log-likelihood) (1912-22)
	MDL/BIC (Minimum description length/Bayesian Information Criterion) (1978)
	AIC (Akaike Information Criterion) (1974)
	NML (Normalized Minimum Likelihood) (2008)
	MIT (Mutual Information Tests) (2006)

References
----------
[1]
http://www.lx.it.pt/~asmc/pub/publications/09-TA/09-c-ta.pdf

"""
from __future__ import division
import numpy as np

from bamt.external.pyBN.utils.independence_tests import mutual_information, entropy


def info_score(bn, data, metric='BIC'):
	if metric.upper() == 'LL':
		score = log_likelihood(bn, data)
	elif metric.upper() == 'BIC':
		score = BIC(bn, data)
	elif metric.upper() == 'AIC':
		score = AIC(bn, data)
	else:
		score = BIC(bn, data)

	return score
	

##### INFORMATION-THEORETIC SCORING FUNCTIONS #####

def log_likelihood(bn, data):
	"""
	Determining log-likelihood of the parameters
	of a Bayesian Network. This is a quite simple
	score/calculation, but it is useful as a straight-forward
	structure learning score.

	Semantically, this can be considered as the evaluation
	of the log-likelihood of the data, given the structure
	and parameters of the BN:
		- log( P( D | Theta_G, G ) )
		where Theta_G are the parameters and G is the structure.

	However, for computational reasons it is best to take
	advantage of the decomposability of the log-likelihood score.
	
	As an example, if you add an edge from A->B, then you simply
	need to calculate LOG(P'(B|A)) - Log(P(B)), and if the value
	is positive then the edge improves the fitness score and should
	therefore be included. 

	Even more, you can expand and manipulate terms to calculate the
	difference between the new graph and the original graph as follows:
		Score(G') - Score(G) = M * I(X,Y),
		where M is the number of data points and I(X,Y) is
		the marginal mutual information calculated using
		the empirical distribution over the data.

	In general, the likelihood score decomposes as follows:
		LL(D | Theta_G, G) = 
			M * Sum over Variables ( I ( X , Parents(X) ) ) - 
			M * Sum over Variables ( H( X ) ),
		where 'I' is mutual information and 'H' is the entropy,
		and M is the number of data points

	Moreover, it is clear to see that H(X) is independent of the choice
	of graph structure (G). Thus, we must only determine the difference
	in the mutual information score of the original graph which had a given
	node and its original parents, and the new graph which has a given node
	and new parents.

	NOTE: This assumes the parameters have already
	been learned for the BN's given structure.

	LL = LL - f(N)*|B|, where f(N) = 0

	Arguments
	---------
	*bn* : a BayesNet object
		Must have both structure and parameters
		instantiated.
	Notes
	-----
	NROW = data.shape[0]
	mi_score = 0
	ent_score = 0
	for rv in bn.nodes():
		cols = tuple([bn.V.index(rv)].extend([bn.V.index(p) for p in bn.parents(rv)]))
		mi_score += mutual_information(data[:,cols])
		ent_score += entropy(data[:,bn.V.index(rv)])
	
	return NROW * (mi_score - ent_score)
	"""

	NROW = data.shape[0]
	mi_score = 0
	ent_score = 0
	for rv in bn.nodes():
		l1 = (bn.V.index(rv),)
		l = tuple([bn.V.index(p) for p in bn.parents(rv)])
		
		cols = l1 + l
		mi_score += mutual_information(data[:,cols])
		ent_score += entropy(data[:,bn.V.index(rv)])
	
	return (NROW * (mi_score - ent_score))
	#return ((1/nrow)*(np.sum(np.log((1e+7+bn.flat_cpt())))))

def MDL(bn, nrow):
	"""
	Minimum Description Length score - it is
	equivalent to BIC
	"""
	return BIC(bn, nrow)

def BIC(bn, data):
	"""
	Bayesian Information Criterion.

	BIC = LL - f(N)*|B|, where f(N) = log(N)/2

	"""
	log_score = log_likelihood(bn, data)
	penalty = 0.5 * bn.num_params() * np.log(max(bn.num_edges(),1))
	return log_score - penalty

def AIC(bn, data):
	"""
	Aikaike Information Criterion

	AIC = LL - f(N)*|B|, where f(N) = 1

	"""
	log_score = log_likelihood(bn, data)
	penalty = len(bn.flat_cpt())
	return log_score - penalty




















"""
*************
Chow-Liu Tree
*************

Calculate the KL divergence (i.e. run
mi_test) between every pair of nodes,
then select the maximum spanning tree from that
connected graph. This is the Chow-Liu tree.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.utils.independence_tests import mi_test
from bamt.external.pyBN.classes.bayesnet import BayesNet
import operator
import numpy as np


def chow_liu(data,edges_only=False):
	"""
	Perform Chow-Liu structure learning algorithm
	over an entire dataset, and return the BN-tree.


	Arguments
	---------
	*data* : a nested numpy array
		The data from which we will learn. It should be
		the entire dataset.

	Returns
	-------
	*bn* : a BayesNet object
		The structure-learned BN.

	Effects
	-------
	None

	Notes
	-----

	"""
	value_dict = dict(zip(range(data.shape[1]),
		[list(np.unique(col)) for col in data.T]))

	n_rv = data.shape[1]

	edge_list = [(i,j,mi_test(data[:,(i,j)],chi2_test=False)) \
					for i in range(n_rv) for j in range(i+1,n_rv)]
	
	edge_list.sort(key=operator.itemgetter(2), reverse=True) # sort by weight
	vertex_cache = {edge_list[0][0]} # start with first vertex..
	mst = dict((rv, []) for rv in range(n_rv))

	for i,j,w in edge_list:
		if i in vertex_cache and j not in vertex_cache:
			mst[i].append(j)
			vertex_cache.add(j)
		elif i not in vertex_cache and j in vertex_cache:
			mst[j].append(i)
			vertex_cache.add(i)
	
	if edges_only == True:
		return mst, value_dict

	bn=BayesNet(mst,value_dict)
	return bn












